"""This module contains various utility methods that support functionality in
other modules.  The multiprocessing functionality also exists in this module,
though that might change with time.

"""

from __future__ import division
import sys, os, time, fnmatch, copy, ctypes
from multiprocessing import Array
from itertools import repeat
from random import shuffle
import datetime

import numpy as np
import nibabel
from scipy.stats import gamma
from scipy.optimize import brute, fmin_powell, fmin
from scipy.stats import linregress
from scipy.integrate import romb, trapz
from scipy import c_, ones, dot, stats, diff
from scipy.linalg import inv, solve, det, norm
from numpy import log, pi, sqrt, square, diagonal
from numpy.random import randn, seed
import sharedmem

# Python 3 compatibility below:
try:  # pragma: no cover
    import cPickle
except ImportError:  # pragma: no cover
    import _pickle as cPickle

try:  # pragma: no cover
    from types import SliceType
except ImportError:  # pragma: no cover
    SliceType = slice

try: # pragma: no cover
    xrange
except NameError:  # pragma: no cover
    xrange = range


def regularizing_error_function(parameter, bundle, p_bounds, thr=0.10): # pragma: no cover
    
    # if out of bounds
    if parameter < p_bounds[0] or parameter > p_bounds[1]:
            return np.inf
    
    # execute
    output = regularizing_objective_function(parameter, bundle)
    
    # what?
    error = np.mean([o.rss for o in output if not np.isnan(o.rss) and o.rsquared > thr])
    
    # when?
    now = str(datetime.datetime.now()).replace(" ","_").replace(".","_").replace(":","_").replace('-','_')
    
    # verbose!
    print('parameter=%.05f    error=%.05f    now=%s' %(o.model.parameter, error, now))
    
    return error

def regularizing_objective_function(parameter, bundle): # pragma: no cover
    
    # attach the guess for tau to each of the voxels in the bundle
    for voxel in bundle:
        model = voxel[1]
        model.parameter = parameter
        
    # fit each of the voxels
    num_cpus = sharedmem.cpu_count()-1
    with sharedmem.Pool(np=num_cpus) as pool:
        output = pool.map(parallel_fit, bundle)
    
    return output

def regularizer(bundle, p_grid, p_bounds, Ns=None): # pragma: no cover
    
    p0 = brute(regularizing_error_function, (p_grid,), args=(bundle, p_bounds), Ns=Ns, finish=None, full_output=True)
    phat = fmin_powell(regularizing_error_function, p0[0], args=(bundle, p_bounds), xtol=1e-4, ftol=1e-4, full_output=True, retall=True)
    
    return p0,phat

def _gamma_difference_hrf(tr, oversampling=1, time_length=32., onset=0.,
                         delay=5, undershoot=15., dispersion=1.,
                         u_dispersion=1., ratio=0.167):
    """ Compute an hrf as the difference of two gamma functions
    Parameters
    ----------
    tr: float, scan repeat time, in seconds
    oversampling: int, temporal oversampling factor, optional
    time_length: float, hrf kernel length, in seconds
    onset: float, onset of the hrf
    Returns
    -------
    hrf: array of shape(length / tr * oversampling, float),
         hrf sampling on the oversampled time grid
    """
    dt = tr / oversampling
    time_stamps = np.linspace(0, time_length, int(float(time_length) / dt))
    time_stamps -= onset / dt
    hrf = gamma.pdf(time_stamps, delay / dispersion, dt / dispersion) - \
        ratio * gamma.pdf(
        time_stamps, undershoot / u_dispersion, dt / u_dispersion)
    hrf /= trapz(hrf)
    return hrf

def spm_hrf(delay, tr, oversampling=1, time_length=32., onset=0.):
    """ Implementation of the SPM hrf model
    Parameters
    ----------
    tr: float, scan repeat time, in seconds
    oversampling: int, temporal oversampling factor, optional
    time_length: float, hrf kernel length, in seconds
    onset: float, onset of the response
    Returns
    -------
    hrf: array of shape(length / tr * oversampling, float),
         hrf sampling on the oversampled time grid
    """
    return _gamma_difference_hrf(tr, oversampling, time_length, onset, delay=5+delay, undershoot=15+delay,)


def glover_hrf(delay, tr, oversampling=1, time_length=32., onset=0.):
    """ Implementation of the Glover hrf model
    Parameters
    ----------
    tr: float, scan repeat time, in seconds
    oversampling: int, temporal oversampling factor, optional
    time_length: float, hrf kernel length, in seconds
    onset: float, onset of the response
    Returns
    -------
    hrf: array of shape(length / tr * oversampling, float),
         hrf sampling on the oversampled time grid
    """
    return _gamma_difference_hrf(tr, oversampling, time_length, onset,
                                delay=5+delay, undershoot=15+delay, dispersion=.9,
                                u_dispersion=.9, ratio=.35)


def grid_slice(start, stop, Ns, dryrun=False):
    
    #### NOTE: there is some weird stuff going on when Ns is => stop-start
    #### it will return an array/slice object that is longer and wide by 1
    #### element. This is easy to fix in the dry-run mode but not so with
    #### the slice object. Why oh why does scipy.brute insist on slice?
    
    # special case
    if Ns == 2:
        step = stop-start
    # all others
    else:
        step = np.diff(np.linspace(start,stop,Ns))[0]
    
    # if true, this return the ndarray rather than slice object.
    if dryrun: # pragma: no cover
        return arange(start, stop+step, step) # pragma: no cover
    else:
        return slice(start, stop+step, step)
        
def distance_mask(x, y, sigma, deg_x, deg_y, amplitude=1):
    
    distance = (deg_x - x)**2 + (deg_y - y)**2
    mask = np.zeros_like(distance, dtype='uint8')
    mask[distance < sigma**2] = 1
    mask *= amplitude
    
    return mask

def recast_estimation_results(output, grid_parent, overloaded=False):
    
    # load the gridParent
    dims = list(grid_parent.shape)
    dims = dims[0:3]
    
    if overloaded == True and output[0].overloaded_estimate is not None:
        dims.append(len(output[0].overloaded_estimate)+1)
    else:
        dims.append(len(output[0].estimate)+1)
        
    # initialize the statmaps
    estimates = np.zeros(dims)
    
    # extract the prf model estimates from the results queue output
    for fit in output:
        
        if not np.isnan(fit.rsquared): # pragma: no cover
        
            # gather the estimate + stats
            if overloaded == True and fit.overloaded_estimate is not None:
                voxel_dat = list(fit.overloaded_estimate)
            else:
                voxel_dat = list(fit.estimate)
                
            voxel_dat.append(fit.rsquared)
            voxel_dat = np.array(voxel_dat)
            
            # assign to
            estimates[tuple(fit.voxel_index)] = voxel_dat
            
    # get header information from the gridParent and update for the prf volume
    aff = grid_parent.get_affine()
    hdr = grid_parent.get_header()
    hdr.set_data_shape(dims)
    
    # recast as nifti
    nifti_estimates = nibabel.Nifti1Image(estimates,aff,header=hdr)
    
    return nifti_estimates

def recast_xval_results(output, bootstraps, indices, grid_parent, overloaded=False, ncpus=30):

    # load the grid_parent (x,y,z)
    dims = list(grid_parent.shape)
    dims = dims[0:3]
    
    # params + rsquared + cod
    if overloaded == True and output[0].overloaded_estimate is not None:
        dims.append(len(output[0].overloaded_estimate)+2)
    else:
        dims.append(len(output[0].estimate)+2)
    
    # add bootstrap dim
    dims.append(bootstraps)
    
    # initialize the statmaps
    estimates = generate_shared_array(np.zeros(dims), ctypes.c_double)
    
    # parallelizer
    def parallel_loader(index): # pragma: no cover
        
        # gather up fits for this voxel
        fits = [o for o in output if list(o.voxel_index) == list(index)]
        
        # gather the estimate + stats
        if overloaded == True and fits[0].overloaded_estimate is not None:
            params = np.array([fit.overloaded_estimate for fit in fits])
        else:
            params = np.array([fit.estimate for fit in fits])
            
        # xval metrics
        rsq = np.array([fit.rsquared for fit in fits])
        cod = np.array([fit.cod for fit in fits])
            
        # assign
        estimates[index[0],index[1],index[2]] = np.concatenate((params, cod[:,np.newaxis], rsq[:,np.newaxis]),-1).T
        
        return None
        
    # populate data structure
    with sharedmem.Pool(np=ncpus) as pool:
        pool.map(parallel_loader, indices)
        
    # header & affine
    aff = grid_parent.get_affine()
    hdr = grid_parent.get_header()
    hdr.set_data_shape(dims)
    
    # recast as nifti
    nifti_estimates = nibabel.Nifti1Image(estimates,aff,header=hdr)
    
    return nifti_estimates

def make_nifti(data, grid_parent=None):
    
    if grid_parent:
        
        # get header information from the gridParent and update for the prf volume
        aff = grid_parent.get_affine()
        hdr = grid_parent.get_header()
        
        # recast as nifti
        nifti = nibabel.Nifti1Image(data,aff,header=hdr)
        
    else:
        aff = np.eye(4,4)
        nifti = nibabel.Nifti1Image(data,aff)
        
    return nifti

def generate_shared_array(unshared_arr,dtype):

    r"""Creates synchronized shared arrays from numpy arrays.

    The function takes a numpy array `unshared_arr` and returns a shared
    memory object, `shared_arr`.  The user also specifies the data-type of
    the values in the array with the `dataType` argument.  See
    multiprocessing.Array and ctypes for details on shared memory arrays and
    the data-types.

    Parameters
    ----------
    unshared_arr : ndarray
        Array_like means all those objects -- lists, nested lists, etc. --
        that can be converted to an array.  We can also refer to
        variables like `var1`.
    dtype : ctypes instance
        The data-type specificed has to be an instance of the ctypes library.
        See ctypes for details.

    Returns
    -------
    shared_arr : synchronized shared array
        An array that is read accessible from multiple processes/threads.
    """

    shared_arr = sharedmem.empty(unshared_arr.shape, dtype=dtype)
    shared_arr[:] = unshared_arr[:]
    return shared_arr

# normalize to a specific range
def normalize(array, imin=-1, imax=1, axis=-1):

    r"""A short-hand function for normalizing an array to a desired range.

    Parameters
    ----------
    array : ndarray
        An array to be normalized.

    imin : float
        The desired minimum value in the output array.  Default: -1

    imax : float
        The desired maximum value in the output array.  Default: 1


    Returns
    -------
    array : ndarray
        The normalized array with imin and imax as the minimum and
        maximum values.
    """

    new_arr = array.copy()

    if np.ndim(imin) == 0:
        dmin = new_arr.min()
        dmax = new_arr.max()
        new_arr = new_arr-dmin
        new_arr = new_arr*(imax - imin)
        new_arr = new_arr/(dmax - dmin)
        new_arr = new_arr+imin
    else:
        dmin = new_arr.min(axis=axis)
        dmax = new_arr.max(axis=axis)
        new_arr -= dmin[:,np.newaxis]
        new_arr *= imax[:,np.newaxis] - imin[:,np.newaxis]
        new_arr /= dmax[:,np.newaxis] - dmin[:,np.newaxis]
        new_arr += imin[:,np.newaxis]

    return new_arr


# generic gradient descent
def gradient_descent_search(data, error_function, objective_function, parameters, bounds, verbose):

    r"""A generic gradient-descent error minimization function.

    The values inside `parameters` are used as a seed-point for a
    gradient-descent error minimization procedure [1]_.  The user must
    also supply an `objective_function` for producing a model prediction
    and an `error_function` for relating that prediction to the actual,
    measured `data`.

    In addition, the user may also supply  `fit_bounds`, containing pairs
    of upper and lower bounds for each of the values in parameters.
    If `fit_bounds` is specified, the error minimization procedure in
    `f` will return an Inf whether the parameters exceed the
    minimum or maxmimum values specified in `fit_bounds`.


    Parameters
    ----------
    parameters : tuple
        A tuple of values representing a model setting.

    args : tuple
        Extra arguments to `objective_function` beyond those in `parameters`.

    fit_bounds : tuple
        A tuple containing the upper and lower bounds for each parameter
        in `parameters`.  If a parameter is not bounded, simply use
        `None`.  For example, `fit_bounds=((0,None),(-10,10),)` would
        bound the first parameter to be any positive number while the
        second parameter would be bounded between -10 and 10.

    data : ndarray
        The actual, measured time-series against which the model is fit.

    error_function : callable
        The error function that relates the model prediction to the
        measure data.  The error function returns a float that represents
        the residual sum of squared errors between the prediction and the
        data.

    objective_function : callable
        The objective function that takes `parameters` and `args` and
        proceduces a model time-series.

    Returns
    -------
    estimate : tuple
        The model solution given `parameters` and `objective_function`.


    References
    ----------

    .. [1] Fletcher, R, Powell, MJD (1963) A rapidly convergent descent
    method for minimization, Compututation Journal 6, 163-168.

    """

    output = fmin_powell(error_function, parameters,
                         args=(bounds, data, objective_function, verbose),
                         full_output=True, disp=False, retall=True)

    return output

def brute_force_search(data, error_function, objective_function, grids, bounds, Ns=None, verbose=False):

    r"""A generic brute-force grid-search error minimization function.

    The user specifies an `objective_function` and the corresponding
    `args` for generating a model prediction.  The `brute_force_search`
    uses `search_bounds` to dictate the bounds for evenly sample the
    parameter space.

    In addition, the user must also supply  `fit_bounds`, containing pairs
    of upper and lower bounds for each of the values in parameters.
    If `fit_bounds` is specified, the error minimization procedure in
    `error_function` will return an Inf whether the parameters exceed the
    minimum or maxmimum values specified in `fit_bounds`.

    The output of `brute_force_search` can be used as a seed-point for
    the fine-tuned solutions from `gradient_descent_search`.


    Parameters
    ----------
    args : tuple
        Arguments to `objective_function` that yield a model prediction.

    grids : tuple
        A tuple indicating the search space for the brute-force grid-search.
        The tuple contains pairs of upper and lower bounds for exploring a
        given dimension.  For example `grids=((-10,10),(0,5),)` will
        search the first dimension from -10 to 10 and the second from 0 to 5.
        These values cannot be `None`.

        For more information, see `scipy.optimize.brute`.

    bounds : tuple
        A tuple containing the upper and lower bounds for each parameter
        in `parameters`.  If a parameter is not bounded, simply use
        `None`.  For example, `fit_bounds=((0,None),(-10,10),)` would
        bound the first parameter to be any positive number while the
        second parameter would be bounded between -10 and 10.

    Ns : int
        Number of samples per stimulus dimension to sample during the ballpark search.

        For more information, see `scipy.optimize.brute`.

    data : ndarray
       The actual, measured time-series against which the model is fit.

    error_function : callable
       The error function that relates the model prediction to the
       measure data.  The error function returns a float that represents
       the residual sum of squared errors between the prediction and the
       data.

    objective_function : callable
      The objective function that takes `parameters` and `args` and
      proceduces a model time-series.

    Returns
    -------
    estimate : tuple
       The model solution given `parameters` and `objective_function`.

    """

    # if user provides their own grids
    if isinstance(grids[0], SliceType):
        output = brute(error_function,
                       args=(bounds, data, objective_function, verbose),
                       ranges=grids,
                       finish=None,
                       full_output=True,
                       disp=False)

    # otherwise specify (min,max) and Ns for each dimension
    else:
        output = brute(error_function,
               args=(bounds, data, objective_function, verbose),
               ranges=grids,
               Ns=Ns,
               finish=None,
               full_output=True,
               disp=False)


    return output

# generic error function
def error_function(parameters, bounds, data, objective_function, verbose):

    r"""A generic error function with bounding.

    Parameters
    ----------
    parameters : tuple
        A tuple of values representing a model setting.

    args : tuple
        Extra arguments to `objective_function` beyond those in `parameters`.

    data : ndarray
       The actual, measured time-series against which the model is fit.

    objective_function : callable
        The objective function that takes `parameters` and `args` and
        proceduces a model time-series.

    debug : bool
        Useful for debugging a model, will print the parameters and error.

    Returns
    -------
    error : float
        The residual sum of squared errors between the prediction and data.
    """
    
    ############
    #   NOTE   #
    ############
    # as of now, this will not work if your model has 1 parameter.
    # i think it is because scipy.optimize.brute returns
    # a scalar when num params is 1, and a tuple/list
    # when num params is > 1. have to look into this further
    
    # check if parameters are inside bounds
    for p, b in zip(parameters,bounds):
        # if not return an inf
        if b[0] and p < b[0]:
            return np.inf
        if b[1] and b[1] < p:
            return np.inf
            
    # merge the parameters and arguments
    ensemble = []
    ensemble.extend(parameters)
    
    # compute the RSS
    prediction = objective_function(*ensemble)
    
    # if nan, return inf
    if np.any(np.isnan(prediction)):
        return np.inf # pragma: no cover
        
    # else, return RSS
    error = np.nansum((data-prediction)**2)
    # error = norm(data-prediction)
    
    # print for debugging
    if verbose:
        print(parameters, error)

    return error

def double_gamma_hrf(delay, tr, fptr=1.0, integrator=trapz):

    r"""The double gamma hemodynamic reponse function (HRF).
    The user specifies only the delay of the peak and undershoot.
    The delay shifts the peak and undershoot by a variable number of
    seconds. The other parameters are hardcoded. The HRF delay is
    modeled for each voxel independently. The form of the HRF and the
    hardcoded values are based on previous work [1]_.

    Parameters
    ----------
    delay : float
        The delay of the HRF peak and undershoot.

    tr : float
        The length of the repetition time in seconds.

    fptr : float
        The number of stimulus frames per reptition time.  For a
        60 Hz projector and with a 1 s repetition time, the fptr
        would be equal to 60.  It is possible that you will bin all
        the frames in a single TR, in which case fptr equals 1.

    integrator : callable
        The integration function for normalizing the units of the HRF
        so that the area under the curve is the same for differently
        delayed HRFs.  Set integrator to None to turn off normalization.

    Returns
    -------
    hrf : ndarray
        The hemodynamic response function to convolve with the stimulus
        timeseries.

    Reference
    ----------
    .. [1] Glover, GH (1999) Deconvolution of impulse response in event related
    BOLD fMRI. NeuroImage 9, 416-429.

    """
    from scipy.special import gamma
    
    # add delay to the peak and undershoot params (alpha 1 and 2)
    alpha_1 = float(5 + delay)
    beta_1 = 1.0
    c = 0.1
    alpha_2 = float(15 + delay)
    beta_2 = 1.0
    
    t = np.arange(0,32,tr)
    
    hrf = ( ( ( t ** (alpha_1) * beta_1 ** alpha_1 * np.exp( -beta_1 * t )) /gamma( alpha_1 )) - c *
            ( ( t ** (alpha_2) * beta_2 ** alpha_2 * np.exp( -beta_2 * t )) /gamma( alpha_2 )) )
            
    if integrator: # pragma: no cover
        hrf /= integrator(hrf)
        
    return hrf

def percent_change(ts, ax=-1):

    r"""Returns the % signal change of each point of the times series
    along a given axis of the array timeseries

    Parameters
    ----------
    ts : ndarray
        an array of time series

    ax : int, optional (default to -1)
        the axis of time_series along which to compute means and stdevs

    Returns
    -------
    ndarray
        the renormalized time series array (in units of %)

    Examples
    --------
    >>> np.set_printoptions(precision=4)  # for doctesting
    >>> ts = np.arange(4*5).reshape(4,5)
    >>> ax = 0
    >>> percent_change(ts,ax)
    array([[-100.    ,  -88.2353,  -78.9474,  -71.4286,  -65.2174],
           [ -33.3333,  -29.4118,  -26.3158,  -23.8095,  -21.7391],
           [  33.3333,   29.4118,   26.3158,   23.8095,   21.7391],
           [ 100.    ,   88.2353,   78.9474,   71.4286,   65.2174]])
    >>> ax = 1
    >>> percent_change(ts,ax)
    array([[-100.    ,  -50.    ,    0.    ,   50.    ,  100.    ],
           [ -28.5714,  -14.2857,    0.    ,   14.2857,   28.5714],
           [ -16.6667,   -8.3333,    0.    ,    8.3333,   16.6667],
           [ -11.7647,   -5.8824,    0.    ,    5.8824,   11.7647]])
    """
    ts = np.asarray(ts)

    return (ts / np.expand_dims(np.mean(ts, ax), ax) - 1) * 100


def zscore(time_series, axis=-1):

    r"""Returns the z-score of each point of the time series
    along a given axis of the array time_series.

    Parameters
    ----------
    time_series : ndarray
        an array of time series
    axis : int, optional
        the axis of time_series along which to compute means and stdevs

    Returns
    _______
    zt : ndarray
        the renormalized time series array
    """

    time_series = np.asarray(time_series)
    et = time_series.mean(axis=axis)
    st = time_series.std(axis=axis)
    sl = [slice(None)] * len(time_series.shape)
    sl[axis] = np.newaxis
    if sl == [None]:
        zt = (time_series - et)/st
    else:
        zt = time_series - et[sl]
        zt /= st[sl]
        
    return zt
    
def bootstrap_bundle(bootstraps, resamples, Fit, model, data, grids, bounds, indices, auto_fit=True, verbose=1, Ns=None):
    
    # initialze
    Fits = []
    
    # main loop
    for resample in resamples:
        for bootstrap in xrange(bootstraps):
            for voxel in xrange(data.shape[0]):
                
                # voxel
                voxel_idx = indices[voxel]
                
                # create random draws
                resample_idx = np.random.choice(np.arange(data.shape[1]-1),resample,replace=False)
                
                # data
                this_data = np.mean(data[voxel,resample_idx,:],0)
                
                # store it
                Fits.append((Fit, model, this_data, grids, bounds, Ns, voxel_idx, auto_fit, verbose, resample_idx))
                
    # randomize list order
    idx = np.argsort(np.random.rand(len(Fits)))
    Fits = [Fits[i] for i in idx]
    
    return Fits

def xval_bundle(bootstraps, kfolds, Fit, model, data, grids, bounds, indices, auto_fit=True, verbose=1, Ns=None):
    
    # num runs
    runs = np.arange(data.shape[1])
    
    # initialize
    Fits = []
    
    # main loop
    for bootstrap in xrange(bootstraps):
        for voxel in xrange(data.shape[0]):
            
            # voxel
            voxel_idx = indices[voxel]
            
            # data
            the_data = data[voxel,:,:]
            
            # create random draws
            if kfolds == 1: # leave one out
                trn_idx = np.random.choice(runs, len(runs)-1, replace=False)
            else:
                trn_idx = np.random.choice(runs, np.int(len(runs)/kfolds), replace=False)
            
            tst_idx = np.array(list(set(runs)-set(trn_idx)))
            
            # compute mean timeseries
            trn_data = np.mean(the_data[trn_idx,:], 0)
            tst_data = np.mean(the_data[tst_idx,:], 0)
            
            # store it
            Fits.append((Fit, model, trn_data, tst_data, grids, bounds, Ns, voxel_idx, auto_fit, verbose, trn_idx, tst_idx))
    
    # randomize list order
    idx = np.argsort(np.random.rand(len(Fits)))
    Fits = [Fits[i] for i in idx]
    
    return Fits

def multiprocess_bundle(Fit, model, data, grids, bounds, indices, auto_fit=True, verbose=1, Ns=None):
    
    # num voxels
    num_voxels = np.int(np.shape(data)[0])
    
    # expand out grids and bounds
    grids = [grids,]*num_voxels
    bounds = [bounds,]*num_voxels
    
    # package the data structure
    dat = zip(repeat(Fit,num_voxels),
              repeat(model,num_voxels),
              data,
              grids,
              bounds,
              repeat(Ns,num_voxels),
              indices,
              repeat(auto_fit,num_voxels),
              repeat(verbose,num_voxels))
    
    # randomize list order
    dat = list(dat)
    idx = np.argsort(np.random.rand(len(dat)))
    dat = [dat[i] for i in idx]
    
    return dat

def gaussian_2D(X, Y, x0, y0, sigma_x, sigma_y, degrees, amplitude=1):
    
    theta = degrees*np.pi/180
    
    a = np.cos(theta)**2/2/sigma_x**2 + np.sin(theta)**2/2/sigma_y**2
    b = -np.sin(2*theta)/4/sigma_x**2 + np.sin(2*theta)/4/sigma_y**2
    c = np.sin(theta)**2/2/sigma_x**2 + np.cos(theta)**2/2/sigma_y**2
    
    Z = amplitude*np.exp( - (a*(X-x0)**2 + 2*b*(X-x0)*(Y-y0) + c*(Y-y0)**2))
    
    return Z

def parallel_xval(args):

    r"""
    This is a convenience function for parallelizing the fitting
    procedure.  Each call is handed a tuple or list containing
    all the necessary inputs for instantiaing a `GaussianFit`
    class object and estimating the model parameters.
    
    
    Paramaters
    ----------
    args : list/tuple
        A list or tuple containing all the necessary inputs for fitting
        the Gaussian pRF model.
        
    Returns
    -------
    
    fit : `Fit` class object
        A fit object that contains all the inputs and outputs of the
        pRF model estimation for a single voxel.
        
    """
    
    # unpackage the arguments
    Fit = args[0]
    model = args[1]
    trn_data = args[2]
    tst_data = args[3]
    grids = args[4]
    bounds = args[5]
    voxel_index = args[6]
    Ns = args[7]
    auto_fit = args[8]
    verbose = args[9]
    trn_idx = args[10]
    tst_idx = args[11]
    
    # fit the data
    fit = Fit(model,
              trn_data,
              grids,
              bounds,
              Ns,
              voxel_index,
              auto_fit,
              verbose)
    
    fit.trn_data = trn_data
    fit.tst_data = tst_data
    fit.trn_idx = trn_idx
    fit.tst_idx = tst_idx
    fit.cod = coeff_of_determination(fit.tst_data, fit.prediction)
    
    return fit
        
def parallel_bootstrap(args):
    
    r"""
    This is a convenience function for parallelizing the fitting
    procedure.  Each call is handed a tuple or list containing
    all the necessary inputs for instantiaing a `GaussianFit`
    class object and estimating the model parameters.
    
    
    Paramaters
    ----------
    args : list/tuple
        A list or tuple containing all the necessary inputs for fitting
        the Gaussian pRF model.
        
    Returns
    -------
    
    fit : `Fit` class object
        A fit object that contains all the inputs and outputs of the
        pRF model estimation for a single voxel.
        
    """
    
    
    # unpackage the arguments
    Fit = args[0]
    model = args[1]
    data = args[2]
    grids = args[3]
    bounds = args[4]
    voxel_index = args[5]
    Ns = args[6]
    auto_fit = args[7]
    verbose = args[8]
    resamples = args[9]
    
    # fit the data
    fit = Fit(model,
              data,
              grids,
              bounds,
              Ns,
              voxel_index,
              auto_fit,
              verbose)
    
    fit.resamples = resamples
    fit.n_resamples = len(resamples)
    
    return fit


def parallel_fit(args):

    r"""
    This is a convenience function for parallelizing the fitting
    procedure.  Each call is handed a tuple or list containing
    all the necessary inputs for instantiaing a `GaussianFit`
    class object and estimating the model parameters.
    
    Paramaters
    ----------
    args : list/tuple
        A list or tuple containing all the necessary inputs for fitting
        the Gaussian pRF model.
        
    Returns
    -------
    
    fit : `Fit` class object
        A fit object that contains all the inputs and outputs of the
        pRF model estimation for a single voxel.
        
    """
    
    
    # unpackage the arguments
    Fit = args[0]
    model = args[1]
    data = args[2]
    grids = args[3]
    bounds = args[4]
    voxel_index = args[5]
    Ns = args[6]
    auto_fit = args[7]
    verbose = args[8]
    
    # fit the data
    fit = Fit(model,
              data,
              grids,
              bounds,
              Ns,
              voxel_index,
              auto_fit,
              verbose)
    return fit

def cartes_to_polar(cartes):

    """
    Assumes that the 0th and 1st parameters are cartesian `x` and `y`
    """
    polar = cartes.copy()
    polar[...,0] = np.mod(np.arctan2(cartes[...,1], cartes[...,0]),2*np.pi)
    polar[...,1] = np.sqrt(cartes[...,0]**2 + cartes[...,1]**2)
    return polar


def find_files(directory, pattern):
    names = []
    for root, dirs, files in os.walk(directory):
        for basename in files:
            if fnmatch.fnmatch(basename, pattern):
                filename = os.path.join(root, basename)
                names.append(filename)
    return names

def binner(signal, times, bins):
    binned_response = np.zeros(len(bins)-2)
    bin_width = bins[1] - bins[0]
    for t in xrange(1,len(bins)):
        the_bin = bins[t]
        binned_signal = signal[(times >= the_bin-bin_width) & (times <= the_bin)]
        binned_response[t-2] = np.sum(binned_signal)
    return binned_response

import sys
from numpy import NaN, Inf, arange, isscalar, asarray, array

def peakdet(v, delta, x = None):
    """
    Converted from MATLAB script at http://billauer.co.il/peakdet.html

    Returns two arrays

    function [maxtab, mintab]=peakdet(v, delta, x)
    #PEAKDET Detect peaks in a vector
    #        [MAXTAB, MINTAB] = PEAKDET(V, DELTA) finds the local
    #        maxima and minima ("peaks") in the vector V.
    #        MAXTAB and MINTAB consists of two columns. Column 1
    #        contains indices in V, and column 2 the found values.
    #
    #        With [MAXTAB, MINTAB] = PEAKDET(V, DELTA, X) the indices
    #        in MAXTAB and MINTAB are replaced with the corresponding
    #        X-values.
    #
    #        A point is considered a maximum peak if it has the maximal
    #        value, and was preceded (to the left) by a value lower by
    #        DELTA.
    #
    # Eli Billauer, 3.4.05 (Explicitly not copyrighted).
    # This function is released to the public domain; Any use is allowed.
    
    """
    maxtab = []
    mintab = []
    
    if x is None: # pragma: no cover
        x = arange(len(v))
        
    v = asarray(v)
    
    mn, mx = Inf, -Inf
    mnpos, mxpos = NaN, NaN
    
    lookformax = True
    
    for i in arange(len(v)):
        this = v[i]
        if this > mx:
            mx = this
            mxpos = x[i]
        if this < mn:
            mn = this
            mnpos = x[i]
            
        if lookformax:
            if this < mx-delta:
                maxtab.append((mxpos, mx))
                mn = this
                mnpos = x[i]
                lookformax = False
        else:
            if this > mn+delta:
                mintab.append((mnpos, mn))
                mx = this
                mxpos = x[i]
                lookformax = True

    return array(maxtab), array(mintab)

def coeff_of_determination(data, model, axis=-1):
    
    r"""
    Calculate the coefficient of determination for a model prediction, relative
    to data.
    
    Parameters
    ----------
    data : ndarray
        The data
    model : ndarray
        The predictions of a model for this data. Same shape as the data.
    axis: int, optional
        The axis along which different samples are laid out (default: -1).
        
    Returns
    -------
    COD : ndarray
       The coefficient of determination.
       
    """
    
    residuals = data - model
    ss_err = np.sum(residuals ** 2, axis=axis)
    
    demeaned_data = data - np.mean(data, axis=axis)[..., np.newaxis]
    ss_tot = np.sum(demeaned_data **2, axis=axis)
    
    # Don't divide by 0:
    if np.all(ss_tot==0.0):
        return np.nan
            
    return 100 * (1 - (ss_err/ss_tot))

# class ols:
#     """
#     Author: Vincent Nijs (+ ?)
#     Email: v-nijs at kellogg.northwestern.edu
#     Last Modified: Mon Jan 15 17:56:17 CST 2007
#     """
#
#     def __init__(self,y,x,y_varnm = 'y',x_varnm = ''):
#
#         self.y = y
#         self.x = c_[ones(x.shape[0]),x]
#         self.y_varnm = y_varnm
#         if not isinstance(x_varnm,list):
#             self.x_varnm = ['const'] + list(x_varnm)
#         else:
#             self.x_varnm = ['const'] + x_varnm
#
#         # Estimate model using OLS
#         self.estimate()
#
#     def estimate(self):
#
#         # estimating coefficients, and basic stats
#         self.inv_xx = inv(dot(self.x.T,self.x))
#         xy = dot(self.x.T,self.y)
#         self.b = dot(self.inv_xx,xy)                    # estimate coefficients
#
#         self.nobs = self.y.shape[0]                     # number of observations
#         self.ncoef = self.x.shape[1]                    # number of coef.
#         self.df_e = self.nobs - self.ncoef              # degrees of freedom, error
#         self.df_r = self.ncoef - 1                      # degrees of freedom, regression
#
#         self.e = self.y - dot(self.x,self.b)            # residuals
#         self.sse = dot(self.e,self.e)/self.df_e         # SSE
#         self.se = sqrt(diagonal(self.sse*self.inv_xx))  # coef. standard errors
#         self.t = self.b / self.se                       # coef. t-statistics
#         self.p = (1-stats.t.cdf(abs(self.t), self.df_e)) * 2    # coef. p-values
#
#         self.R2 = 1 - self.e.var()/self.y.var()         # model R-squared
#         self.R2adj = 1-(1-self.R2)*((self.nobs-1)/(self.nobs-self.ncoef))   # adjusted R-square
#
#         self.F = (self.R2/self.df_r) / ((1-self.R2)/self.df_e)  # model F-statistic
#         self.Fpv = 1-stats.f.cdf(self.F, self.df_r, self.df_e)  # F-statistic p-value
#
#     def dw(self):
#         """
#         Calculates the Durbin-Waston statistic
#         """
#         de = diff(self.e,1)
#         dw = dot(de,de) / dot(self.e,self.e);
#
#         return dw
#
#     def omni(self):
#         """
#         Omnibus test for normality
#         """
#         return stats.normaltest(self.e)
#
#     def JB(self):
#         """
#         Calculate residual skewness, kurtosis, and do the JB test for normality
#         """
#
#         # Calculate residual skewness and kurtosis
#         skew = stats.skew(self.e)
#         kurtosis = 3 + stats.kurtosis(self.e)
#
#         # Calculate the Jarque-Bera test for normality
#         JB = (self.nobs/6) * (square(skew) + (1/4)*square(kurtosis-3))
#         JBpv = 1-stats.chi2.cdf(JB,2);
#
#         return JB, JBpv, skew, kurtosis
#
#     def ll(self):
#         """
#         Calculate model log-likelihood and two information criteria
#         """
#
#         # Model log-likelihood, AIC, and BIC criterion values
#         ll = -(self.nobs*1/2)*(1+log(2*pi)) - (self.nobs/2)*log(dot(self.e,self.e)/self.nobs)
#         aic = -2*ll/self.nobs + (2*self.ncoef/self.nobs)
#         bic = -2*ll/self.nobs + (self.ncoef*log(self.nobs))/self.nobs
#
#         return ll, aic, bic
#
#     def summary(self):
#         """
#         Printing model output to screen
#         """
#
#         # local time & date
#         t = time.localtime()
#
#         # extra stats
#         ll, aic, bic = self.ll()
#         JB, JBpv, skew, kurtosis = self.JB()
#         omni, omnipv = self.omni()
#
#         # printing output to screen
#         print '\n=============================================================================='
#         print "Dependent Variable: " + self.y_varnm
#         print "Method: Least Squares"
#         print "Date: ", time.strftime("%a, %d %b %Y",t)
#         print "Time: ", time.strftime("%H:%M:%S",t)
#         print '# obs:               %5.0f' % self.nobs
#         print '# variables:     %5.0f' % self.ncoef
#         print '=============================================================================='
#         print 'variable     coefficient     std. Error      t-statistic     prob.'
#         print '=============================================================================='
#         for i in range(len(self.x_varnm)):
#             print '''% -5s          % -5.6f     % -5.6f     % -5.6f     % -5.6f''' % tuple([self.x_varnm[i],self.b[i],self.se[i],self.t[i],self.p[i]])
#         print '=============================================================================='
#         print 'Models stats                         Residual stats'
#         print '=============================================================================='
#         print 'R-squared            % -5.6f         Durbin-Watson stat  % -5.6f' % tuple([self.R2, self.dw()])
#         print 'Adjusted R-squared   % -5.6f         Omnibus stat        % -5.6f' % tuple([self.R2adj, omni])
#         print 'F-statistic          % -5.6f         Prob(Omnibus stat)  % -5.6f' % tuple([self.F, omnipv])
#         print 'Prob (F-statistic)   % -5.6f           JB stat             % -5.6f' % tuple([self.Fpv, JB])
#         print 'Log likelihood       % -5.6f           Prob(JB)            % -5.6f' % tuple([ll, JBpv])
#         print 'AIC criterion        % -5.6f         Skew                % -5.6f' % tuple([aic, skew])
#         print 'BIC criterion        % -5.6f         Kurtosis            % -5.6f' % tuple([bic, kurtosis])
#         print '=============================================================================='



#!/usr/bin/python

""" Classes and functions for fitting population encoding models """

from __future__ import division
import warnings
warnings.simplefilter("ignore")

import numpy as np
from scipy.signal import fftconvolve
import nibabel

from popeye.onetime import auto_attr
import popeye.utilities as utils
from popeye.base import PopulationModel, PopulationFit
from popeye.spinach import generate_gabor_receptive_field, generate_rf_timeseries

class GaborModel(PopulationModel):
    
    r"""
    Gabor population receptive field model.
    
    """
    
    def __init__(self, stimulus, hrf_model):
        
        # this is a weird notation
        PopulationModel.__init__(self, stimulus, hrf_model)
        
        
    def generate_ballpark_prediction(self, x, y, sigma, theta, phi, cpd):
        
        # make sure theta and phi are 0-2*pi
        theta = np.mod(theta, 2*np.pi)
        phi = np.mod(theta, 2*np.pi)
        
        # create mask for speed
        distance = (self.stimulus.deg_x0 - x)**2 + (self.stimulus.deg_y0 - y)**2
        mask = np.zeros_like(distance, dtype='uint8')
        mask[distance < (5*sigma)**2] = 1
        
        # generate the RF
        rf = generate_gabor_receptive_field(x, y, sigma, theta, phi, cpd,
                                            self.stimulus.deg_x0,
                                            self.stimulus.deg_y0)
        
        # extract the stimulus time-series
        response = generate_rf_timeseries(self.stimulus.stim_arr0, rf, mask)
        
        # convolve with the HRF
        hrf = self.hrf_model(0, self.stimulus.tr_length)
        
        # convolve it with the stimulus
        model = fftconvolve(response, hrf, 'same')
        
        return model
    
    
    def generate_prediction(self, x, y, sigma, theta, phi, cpd):
        
        # make sure theta and phi are 0-2*pi
        theta = np.mod(theta, 2*np.pi)
        phi = np.mod(theta, 2*np.pi)
        
        # create mask for speed
        distance = (self.stimulus.deg_x - x)**2 + (self.stimulus.deg_y - y)**2
        mask = np.zeros_like(distance, dtype='uint8')
        mask[distance < (5*sigma)**2] = 1
        
        # generate the RF
        rf = generate_gabor_receptive_field(x, y, sigma, theta, phi, cpd,
                                            self.stimulus.deg_x, self.stimulus.deg_y)
        
        # extract the stimulus time-series
        response = generate_rf_timeseries(self.stimulus.stim_arr, rf, mask)
        
        # convolve with the HRF
        hrf = self.hrf_model(0, self.stimulus.tr_length)
        
        # convolve it with the stimulus
        model = fftconvolve(response, hrf, 'same')
        
        return model
            
class GaborFit(PopulationFit):
    
    r"""
    A class containing tools for fitting the 1D Gaussian pRF model.
    
    """
    
    def __init__(self, model, data, grids, bounds, Ns,
                 voxel_index=(1,2,3), auto_fit=True, verbose=0):
                 
        r"""
        A class containing tools for fitting the Difference of Gaussians pRF model.
        
        The `CompressiveSpatialSummationFit` class houses all the fitting tool that 
        are associated with estimating a pRF model. The `GaussianFit` takes a 
        `CompressiveSpatialSummationModel` instance  `model` and a time-series `data`. 
        In addition, extent and sampling-rate of a  brute-force grid-search is set 
        with `grids` and `Ns`.  Use `bounds` to set limits on the search space for 
        each parameter.  
        
        Paramaters
        ----------
        
                
        model : `CompressiveSpatialSummationModel` class instance
            An object representing the CSS model. 
        
        data : ndarray
            An array containing the measured BOLD signal of a single voxel.
        
        grids : tuple
            A tuple indicating the search space for the brute-force grid-search.
            The tuple contains pairs of upper and lower bounds for exploring a
            given dimension.  For example `grids=((-10,10),(0,5),)` will
            search the first dimension from -10 to 10 and the second from 0 to 5.
            These values cannot be `None`. 
            
            For more information, see `scipy.optimize.brute`.
        
        bounds : tuple
            A tuple containing the upper and lower bounds for each parameter
            in `parameters`.  If a parameter is not bounded, simply use
            `None`.  For example, `fit_bounds=((0,None),(-10,10),)` would 
            bound the first parameter to be any positive number while the
            second parameter would be bounded between -10 and 10.
        
        Ns : int
            Number of samples per stimulus dimension to sample during the ballpark search.
            
            For more information, see `scipy.optimize.brute`.
        
        voxel_index : tuple
            A tuple containing the index of the voxel being modeled. The 
            fitting procedure does not require a voxel index, but 
            collating the results across many voxels will does require voxel
            indices. With voxel indices, the brain volume can be reconstructed 
            using the newly computed model estimates.
        
        auto_fit : bool
            A flag for automatically running the fitting procedures once the 
            `GaussianFit` object is instantiated.
        
        verbose : int
            0 = silent
            1 = print the final solution of an error-minimization
            2 = print each error-minimization step
        
        """
        
        PopulationFit.__init__(self, model, data, grids, bounds, Ns, 
                               voxel_index, auto_fit, verbose)
    @auto_attr
    def x0(self):
        return self.ballpark[0]
        
    @auto_attr
    def y0(self):
        return self.ballpark[1]
    
    @auto_attr
    def s0(self):
        return self.ballpark[2]
        
    @auto_attr
    def hrf0(self):
        return self.ballpark[3]
    
    @auto_attr
    def theta0(self):
        return self.ballpark[4]
    
    @auto_attr
    def phi0(self):
        return self.ballpark[5]
            
    @auto_attr
    def cpd0(self):
        return self.ballpark[6]
            
    @auto_attr
    def x(self):
        return self.estimate[0]
                
    @auto_attr
    def y(self):
        return self.estimate[1]
            
    @auto_attr
    def sigma(self):
        return self.estimate[2]
        
    @auto_attr
    def hrf_delay(self):
        return self.estimate[3]
    
    @auto_attr
    def theta(self):
        return self.estimate[4]
        
    @auto_attr
    def phi(self):
        return self.estimate[5]
            
    @auto_attr
    def cpd(self):
        return self.estimate[6]
        
    @auto_attr
    def receptive_field(self):
        return generate_gabor_receptive_field(*self.estimate)
    



#!/usr/bin/python

""" Classes and functions for fitting population encoding models """

from __future__ import division
import time
import warnings
import gc
warnings.simplefilter("ignore")

import numpy as np
from scipy.optimize import brute, fmin_powell
from scipy.special import gamma
from scipy.stats import linregress
from scipy.signal import fftconvolve, decimate
from scipy.ndimage.filters import median_filter
from scipy.integrate import romb, trapz
from scipy.interpolate import interp1d
import nibabel

from popeye.onetime import auto_attr
import popeye.utilities as utils
from popeye.base import PopulationModel, PopulationFit
from popeye.spinach import generate_rf_timeseries_1D

def recast_estimation_results(output, grid_parent, write=True):
    
    # load the gridParent
    dims = list(grid_parent.shape)
    dims = dims[0:3]
    dims.append(5)
    
    # initialize the statmaps
    nii_out = np.zeros(dims)
    
    # extract the pRF model estimates from the results queue output
    for fit in output:
        if fit.__dict__.has_key('OLS'):
            nii_out[fit.voxel_index] = (fit.center_freq, 
                                        fit.sigma,
                                        fit.rsquared,
                                        fit.coefficient,
                                        fit.stderr)
                                 
    # get header information from the gridParent and update for the pRF volume
    aff = grid_parent.get_affine()
    hdr = grid_parent.get_header()
    hdr.set_data_shape(dims)
    
    # recast as nifti
    nif = nibabel.Nifti1Image(nii_out,aff,header=hdr)
    
    return nif

# this is actually not used, but it serves as the model for the cython function ...
def gaussian_2D(X, Y, x0, y0, sigma_x, sigma_y, degrees, amplitude=1):
    
    theta = degrees*np.pi/180
        
    a = np.cos(theta)**2/2/sigma_x**2 + np.sin(theta)**2/2/sigma_y**2
    b = -np.sin(2*theta)/4/sigma_x**2 + np.sin(2*theta)/4/sigma_y**2
    c = np.sin(theta)**2/2/sigma_x**2 + np.cos(theta)**2/2/sigma_y**2
    
    Z = amplitude*np.exp( - (a*(X-x0)**2 + 2*b*(X-x0)*(Y-y0) + c*(Y-y0)**2))
    
    return Z

def gaussian_1D(freqs, center_freq, sigma):
    
    gaussian = np.exp(-0.5 * ((freqs - center_freq)/sigma)**2)
    
    return gaussian

def compute_model_ts(center_freq, sigma,
                     spectrogram, freqs, target_times):
    
    # generate stimulus time-series
    rf = gaussian_1D(freqs, center_freq, sigma)
    # make sure sigma isn't too big
    # if np.any(np.round(rf[0],3) > 0):
    #     return np.inf
    # if np.any(np.round(rf[-1],3) > 0):
    #     return np.inf
    
    # create mask for speed
    distance = freqs - center_freq
    mask = np.zeros_like(distance, dtype='uint8')
    mask[distance < (5*sigma)] = 1
        
    # extract the response
    stim = generate_rf_timeseries_1D(spectrogram,rf,mask)
    
    # recast the stimulus into a time-series that i can 
    source_times = np.linspace(0,target_times[-1],len(stim),endpoint=True)
    f = interp1d(source_times,stim,kind='linear')
    new_stim = f(target_times)
    
    # hard-set the hrf_delay
    hrf_delay = 0
    
    # convolve it with the HRF
    hrf = utils.double_gamma_hrf(hrf_delay, 1.0, 10)
    stim_pad = np.tile(new_stim,3)
    model = fftconvolve(stim_pad, hrf,'same')[len(new_stim):len(new_stim)*2]
    
    # normalize it
    model = utils.zscore(model)
    
    return model

# this method is used to simply multiprocessing.Pool interactions
def parallel_fit(args):
    
    # unpackage the arguments
    model = args[0]
    data = args[1]
    grids = args[2]
    bounds = args[3]
    Ns = args[4]
    tr_length = args[5]
    voxel_index = args[6]
    auto_fit = args[7]
    verbose = args[8]
    
    # fit the data
    fit = SpectroTemporalFit(model,
                             data,
                             grids,
                             bounds,
                             Ns,
                             tr_length,
                             voxel_index,
                             auto_fit,
                             verbose)
    return fit


class SpectroTemporalModel(PopulationModel):
    
    """
    Gaussian population receptive field model.
    """
    
    def __init__(self, stimulus):
        
        # this is a weird notation
        PopulationModel.__init__(self, stimulus)


class SpectroTemporalFit(PopulationFit):
    
    def __init__(self, model, data, grids, bounds, Ns, tr_length,
                 voxel_index=(1,2,3), auto_fit=True, verbose=0):
        
        
        PopulationFit.__init__(self, model, data, grids, bounds, Ns, tr_length, voxel_index, auto_fit, verbose)
        
        if self.auto_fit:
            
            self.start = time.clock()
            try:
                self.ballpark;
                self.estimate;
                self.OLS;
                self.finish = time.clock()
                
                if self.verbose:
                    print(self.msg)
                    
            except:
                self.finish = time.clock()
                if self.verbose:
                    print(self.errmsg)
    
    @auto_attr
    def ballpark(self):
        return utils.brute_force_search((self.model.stimulus.spectrogram,
                                         self.model.stimulus.freqs,
                                         self.model.stimulus.target_times),
                                        self.grids,
                                        self.bounds,
                                        self.Ns,
                                        self.data,
                                        utils.error_function,
                                        compute_model_ts,
                                        self.very_verbose)
                                        
    @auto_attr
    def estimate(self):
        return utils.gradient_descent_search((self.center_freq0, self.sigma0),
                                             (self.model.stimulus.spectrogram,
                                              self.model.stimulus.freqs,
                                              self.model.stimulus.target_times),
                                             self.bounds,
                                             self.data,
                                             utils.error_function,
                                             compute_model_ts,
                                             self.very_verbose)
                                             
    @auto_attr
    def center_freq0(self):
        return self.ballpark[0]
    
    @auto_attr
    def sigma0(self):
        return self.ballpark[1]
    
    # @auto_attr
    # def beta0(self):
    #     return self.ballpark[2]
    # 
    # @auto_attr
    # def baseline0(self):
    #     return self.ballpark[3]
    
    @auto_attr
    def center_freq(self):
        return self.estimate[0]
        
    @auto_attr
    def sigma(self):
        return self.estimate[1]
    
    # @auto_attr
    # def beta(self):
    #     return self.estimate[2]
    # 
    # 
    # @auto_attr
    # def baseline(self):
    #     return self.estimate[3]
    
    @auto_attr
    def prediction(self):
        return compute_model_ts(self.center_freq, self.sigma,
                                self.model.stimulus.spectrogram,
                                self.model.stimulus.freqs,
                                self.model.stimulus.target_times)
    
    @auto_attr
    def OLS(self):
        return sm.OLS(self.data,self.prediction).fit()
    
    @auto_attr
    def coefficient(self):
        return self.OLS.params[0]
    
    @auto_attr
    def rsquared(self):
        return self.OLS.rsquared
    
    @auto_attr
    def stderr(self):
        return np.sqrt(self.OLS.mse_resid)
    
    @auto_attr
    def rss(self):
        return np.sum((self.data - self.prediction)**2)
    
    @auto_attr
    def receptive_field(self):
        return gaussian_1D(self.model.stimulus.freqs, self.center_freq,)
    
    @auto_attr
    def msg(self):
        txt = ("VOXEL=(%.03d,%.03d,%.03d)   TIME=%.03d   RSQ=%.02f  CENTER=%.02f   SIGMA=%.02f" 
            %(self.voxel_index[0],
              self.voxel_index[1],
              self.voxel_index[2],
              self.finish-self.start,
              self.rsquared,
              self.center_freq,
              self.sigma))
        return txt
    
    @auto_attr
    def errmsg(self):
        txt = ("VOXEL=(%.03d,%.03d,%.03d)   TIME=%.03d   There was a problem with this voxel!" 
            %(self.voxel_index[0],
              self.voxel_index[1],
              self.voxel_index[2],
              self.finish-self.start))
        return txt
        
                                    



"""
"""

__docformat__ = 'restructuredtext'

from .version import  __version__

from . import visual_stimulus
from . import utilities
from . import og
from . import css
from . import dog
from . import xvalidation
from . import plotting



from __future__ import division

import os, ctypes

import popeye.utilities as utils
import numpy as np
import numpy.testing as npt

import nose.tools as nt

from popeye.visual_stimulus import pixels_per_degree, generate_coordinate_matrices, resample_stimulus, simulate_sinflicker_bar, simulate_bar_stimulus, VisualStimulus


def test_pixels_per_degree():
    
    pixels_across = 1
    screen_width = 1
    viewing_distance = 1
    ppd = pixels_per_degree(pixels_across, screen_width, viewing_distance)
    
def test_generate_coordinate_matrices():
    
    # set some dummy display parameters
    pixels_across = 100
    pixels_down = 100
    ppd = 1.0
    scale_factor = 1.0
    
    # generate coordinates
    deg_x, deg_y = generate_coordinate_matrices(pixels_across,pixels_down,ppd,scale_factor)
    
    # assert
    nt.assert_true(np.sum(deg_x[0,0:50]) == np.sum(deg_x[0,50::])*-1)
    nt.assert_true(np.sum(deg_y[0:50,0]) == np.sum(deg_y[50::,0])*-1)
    
    # try the same with an odd number of pixels
    pixels_across = 101
    pixels_down = 101
    
    # generate coordinates
    deg_x, deg_y = generate_coordinate_matrices(pixels_across,pixels_down,ppd,scale_factor)
    
    # assert
    nt.assert_true(np.sum(deg_x[0,0:50]) == np.sum(deg_x[0,50::])*-1)
    nt.assert_true(np.sum(deg_y[0:50,0]) == np.sum(deg_y[50::,0])*-1)
    
    # try with another rescaling factor
    scale_factor = 0.5
    
    # get the horizontal and vertical coordinate matrices
    deg_x, deg_y = generate_coordinate_matrices(pixels_across,pixels_down,ppd,scale_factor)
    
    # assert
    nt.assert_true(np.sum(deg_x[0,0:50]) == np.sum(deg_x[0,50::])*-1)
    nt.assert_true(np.sum(deg_y[0:50,0]) == np.sum(deg_y[50::,0])*-1)

def test_noresample_stimulus():
    
    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,90)
    num_blank_steps = 0
    num_bar_steps = 30
    ecc = 12
    tr_length = 1.0
    frames_per_tr = 1.0
    scale_factor = 1.0
    pixels_across = 100
    pixels_down = 100
    dtype = ctypes.c_int16
    Ns = 3
    voxel_index = (1,2,3)
    auto_fit = True
    verbose = 1
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, 
                                screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # stimulus
    npt.assert_equal(stimulus.stim_arr.shape[0:2],stimulus.stim_arr0.shape[0:2])
    npt.assert_equal(stimulus.deg_x.shape,stimulus.deg_x0.shape)
    npt.assert_equal(stimulus.deg_y.shape,stimulus.deg_y0.shape)

def test_resample_stimulus():
    
    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,90)
    num_blank_steps = 0
    num_bar_steps = 30
    ecc = 12
    tr_length = 1.0
    frames_per_tr = 1.0
    scale_factor = 0.50
    pixels_across = 100
    pixels_down = 100
    dtype = ctypes.c_int16
    Ns = 3
    voxel_index = (1,2,3)
    auto_fit = True
    verbose = 1
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, 
                                screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # grab the stimulus dimensions
    stim_dims = stimulus.stim_arr.shape
    stim_coarse_dims = stimulus.stim_arr0.shape
    
    # assert
    nt.assert_true(stim_coarse_dims[0]/stim_dims[0] == scale_factor)
    nt.assert_true(stim_coarse_dims[1]/stim_dims[1] == scale_factor)
    nt.assert_true(stim_coarse_dims[2] == stim_dims[2])
    
    npt.assert_array_equal(np.unique(stimulus.stim_arr0), [0, 1])
    
    # make sure the duty-cycle doesn't change with resampling
    npt.assert_almost_equal(np.sum(stimulus.stim_arr0==1)/np.sum(stimulus.stim_arr0>-1),np.sum(stimulus.stim_arr==1)/np.sum(stimulus.stim_arr>-1),3)

def test_simulate_sinflicker_bar():
    
    # no blanks
    bar = simulate_sinflicker_bar(100,100,50,10,[0],1,10,5,1,1,60)
    y = utils.normalize(bar[1,1,:],0,1)
    t = np.linspace(0,1,60)
    yhat = utils.normalize(np.sin(2 * np.pi * t),0,1)
    nt.assert_almost_equal(np.sum(yhat-y),0)
    
    # blanks
    bar = simulate_sinflicker_bar(100,100,50,10,[0,-1],1,10,5,1,1,60)
    y = np.round(utils.normalize(bar[1,1,:],0,1),2)
    t = np.linspace(0,1,60)
    yhat = utils.normalize(np.sin(2 * np.pi * t),0,1)
    yhat = np.append(yhat,np.repeat(0.5,60))
    nt.assert_almost_equal(np.sum(yhat-np.round(y,2)),0,1)
    



from __future__ import division
import time
import ctypes
import shutil
from multiprocessing import Process, Queue, Array

import numpy as np
from scipy.interpolate import interp1d

import nibabel

def reconstruct_stimulus(metaData,stimData,funcData,verbose=True):
    
    # grab voxel indices
    xi,yi,zi = metaData['core_voxels']
    
    # set up the time vector for interpolation of the time-series based on the
    # HRF tau estimate 
    runLength = np.shape(stimData['stimRecon'])[-1]
    usTime = np.array([np.round(item,1) for item in np.r_[0:runLength-1:0.1]])
    
    # printing niceties
    numVoxels = len(xi)
    voxelCount = 1
    printLength = len(xi)/10
    
    # printing niceties
    numVoxels = len(xi)
    voxelCount = 1
    printLength = len(xi)/10
    
    for xvoxel,yvoxel,zvoxel in zip(xi,yi,zi):
        
        # Grab timestamp
        toc = time.clock()
        
        # grab the pRF estimate for this voxel
        pRFx,pRFy,pRFs,pRFd = funcData['pRF_polar'][xvoxel,yvoxel,zvoxel,0:4]
        pRFcov = funcData['pRF_polar'][xvoxel,yvoxel,zvoxel,6]
        
        # create the receptive field from the metaData['pRF_polar'] estimate
        rf = MakeFastRF(stimData['degXFine'],stimData['degYFine'],pRFx,pRFy,pRFs)

        # grab the voxel's time-series
        tsActual = funcData['bold'][xvoxel,yvoxel,zvoxel,:]
        
        # create the upsampled time-series
        f = interp1d(np.arange(runLength), tsActual, kind='cubic')
        usActual = f(usTime)

        for tr in range(runLength):
            usTimePoint = np.nonzero(usTime == np.round(tr+pRFd+5,1))[0]
            if usTimePoint < runLength*10:
                intensity = usActual[usTimePoint][0]
                stimData['stimRecon'][:,:,tr] += intensity*rf
        
        # Grab a timestamp
        tic = time.clock()
        
        if verbose:
            percentDone = (voxelCount/numVoxels)*100
            print("%.02d%%  VOXEL=(%.03d,%.03d,%.03d)  TIME=%.03E" 
                  %(percentDone,
                    xvoxel,
                    yvoxel,
                    zvoxel,
                    tic-toc))
                    
    return None
    
def reconstruct_stimulus_realtime(voxels,stimData,funcData,pRFs,verbose=True):
    
    # grab voxel indices
    xi,yi,zi = voxels[:]
    
    # printing niceties
    numVoxels = len(xi)
    voxelCount = 1
    
    # initialize an empty frame to store the reconstruction
    emptyFrame = np.zeros_like(stimData['stimArrayFine'][:,:,0])
    
    for voxel in range(numVoxels):
        
        # grab voxel coordinate
        xvoxel,yvoxel,zvoxel = voxels[0][voxel],voxels[1][voxel],voxels[2][voxel]
        
        # grab the RF from the pRFs array
        rf = pRFs[:,:,voxel]
        
        # smooth the time-series
        intensity = funcData[xvoxel,yvoxel,zvoxel]
        
        # create the receptive field from the pRF estimatetimate
        emptyFrame += rf*intensity
        
        if verbose:
            print("VOXEL=(%.03d,%.03d,%.03d)" %(xvoxel,yvoxel,zvoxel))
            
    return emptyFrame
    
def reconstruct_stimulus_realtime_smoothing(voxels,stimData,funcData,pRFs,verbose=True):

    # grab voxel indices
    xi,yi,zi = voxels[:]

    # printing niceties
    numVoxels = len(xi)
    voxelCount = 1

    # initialize an empty frame to store the reconstruction
    emptyFrame = np.zeros_like(stimData['stimArrayFine'][:,:,0])

    for voxel in range(numVoxels):

        # grab voxel coordinate
        xvoxel,yvoxel,zvoxel = voxels[0][voxel],voxels[1][voxel],voxels[2][voxel]

        # grab the RF from the pRFs array
        rf = pRFs[:,:,voxel]

        # smooth the time-series
        ts = funcData[xvoxel,yvoxel,zvoxel,:]
        intensity = wiener(ts,5)[-1]

        # create the receptive field from the pRF estimate
        emptyFrame += rf*intensity

        if verbose:
            print("VOXEL=(%.03d,%.03d,%.03d,%.03f,%.03f,%.03f)" %(xvoxel,yvoxel,zvoxel))

    return emptyFrame

def multiprocess_stimulus_reconstruction(stimData,funcData,metaData):

    # figure out how many voxels are in the mask & the number of jobs we have allocated
    [xi,yi,zi] = metaData['voxels']
    cpus = metaData['cpus']
    
    # Set up the voxel lists for each job
    voxelLists = []
    cutOffs = [int(np.floor(i)) for i in np.linspace(0,len(xi),cpus+1)]
    for i in range(len(cutOffs)-1):
        l = range(cutOffs[i],cutOffs[i+1])
        voxelLists.append(l)
        
    # start the jobs
    procs = []
    for j in range(cpus):
        voxels = [xi[voxelLists[j]],yi[voxelLists[j]],zi[voxelLists[j]]]
        metaData['core_voxels'] = voxels
        p = Process(target=reconstruct_stimulus,args=(metaData,stimData,funcData))
        procs.append(p)
        p.start()
        
    # close the jobs
    for p in procs:
        p.join()
        
    return None
    

def multiprocess_stimulus_reconstruction_realtime(stimData,funcData,metaData,
                                                  pRF,tr):
    
    # figure out how many voxels are in the mask & the number of jobs we have
    # allocated s
    [xi,yi,zi] = metaData['voxels']
    cpus = metaData['cpus']
    
    # Set up the voxel lists for each job
    voxelLists = []
    cutOffs = [int(np.floor(i)) for i in np.linspace(0,len(xi),cpus+1)]
    for i in range(len(cutOffs)-1):
        l = range(cutOffs[i],cutOffs[i+1])
        voxelLists.append(l)

    # start the jobs
    procs = []
    for j in range(cpus):
        voxels = [xi[voxelLists[j]],yi[voxelLists[j]],zi[voxelLists[j]]]
        metaData['core_voxels'] = voxels
        p = Process(target=reconstruct_stimulus_realtime,args=(metaData,stimData,funcData,pRF,tr))
        procs.append(p)
        p.start()

    # close the jobs
    for p in procs:
        p.join()

    return None



cimport cython
from cython.parallel import prange, parallel, threadid
import numpy as np
cimport numpy as np
from scipy.signal.sigtools import _correlateND
import ctypes

DTYPE = np.uint8
ctypedef np.uint8_t DTYPE_t

DTYPE2 = np.double
ctypedef np.double_t DTYPE2_t

DTYPE3 = np.short
ctypedef short DTYPE3_t

from libc.math cimport sin, cos, exp, sqrt

try:
    xrange
except NameError:  # python3
    xrange = range


@cython.boundscheck(False)
@cython.wraparound(False)
def binner(np.ndarray[DTYPE2_t, ndim=1] signal,
           np.ndarray[DTYPE2_t, ndim=1] times,
           np.ndarray[DTYPE2_t, ndim=1] bins):

    # type defs
    cdef int t, t_lim
    cdef DTYPE2_t bin_width

    # vars
    t_lim = len(bins)
    bin_width = bins[1] - bins[0]

    # output
    cdef np.ndarray[DTYPE2_t, ndim=1, mode='c'] binned_response = np.zeros((t_lim)-2)

    for t in xrange(1,t_lim):
        the_bin = bins[t]
        binned_signal = signal[(times >= the_bin-bin_width) & (times <= the_bin)]
        binned_response[t-2] = np.sum(binned_signal)

    return binned_response

@cython.boundscheck(False)
@cython.wraparound(False)
def two_dimensional_og(np.ndarray[DTYPE2_t, ndim=2] deg_x,
                       np.ndarray[DTYPE2_t, ndim=2] deg_y,
                       DTYPE2_t x,
                       DTYPE2_t y,
                       DTYPE2_t sigma_x,
                       DTYPE2_t sigma_y,
                       DTYPE2_t degrees,
                       DTYPE2_t amplitude):

    # cdef's
    cdef int i,j,k
    cdef DTYPE2_t s_factor2
    cdef DTYPE2_t s_factor3
    cdef int xlim = deg_x.shape[0]
    cdef int ylim = deg_y.shape[1]
    cdef DTYPE2_t d

    # convert degress to radians
    cdef DTYPE2_t theta = degrees * np.pi/180

    # figure out which sigma is bigger and limit our search to within that
    if sigma_x > sigma_y:
        s_factor2 = (2.0*sigma_x**2)
        s_factor3 = (3.0*sigma_x)**2
    else:
        s_factor2 = (2.0*sigma_y**2)
        s_factor3 = (3.0*sigma_y)**2

    # set up the transformation matrix
    cdef DTYPE2_t a = np.cos(theta)**2/2/sigma_x**2 + np.sin(theta)**2/2/sigma_y**2
    cdef DTYPE2_t b = -np.sin(2*theta)/4/sigma_x**2 + np.sin(2*theta)/4/sigma_y**2
    cdef DTYPE2_t c = np.sin(theta)**2/2/sigma_x**2 + np.cos(theta)**2/2/sigma_y**2

    # initialize output variable
    cdef np.ndarray[DTYPE2_t, ndim=2, mode='c'] rf = np.zeros((xlim,ylim),dtype=DTYPE2)

    # the loop
    for i in xrange(xlim):
        for j in xrange(ylim):
            d = (deg_x[i,j]-x)**2 + (deg_y[i,j]-y)**2
            if d <= s_factor3:
                rf[i,j] = np.exp( - (a*(deg_x[i,j]-x)**2 + 2*b*(deg_x[i,j]-x)*(deg_y[i,j]-y) + c*(deg_y[i,j]-y)**2))

    # return it
    return rf

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_mp_timeseries(np.ndarray[DTYPE2_t, ndim=1] spatial_ts,
                           np.ndarray[DTYPE2_t, ndim=1] m_amp,
                           np.ndarray[DTYPE2_t, ndim=1] p_amp,
                           np.ndarray[DTYPE_t, ndim=1] flicker_vec):
    # cdef's
    cdef int t
    cdef int tlim = spatial_ts.shape[0]
    
    # initialize output variable
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] m_ts = np.zeros(tlim,dtype=DTYPE2)
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] p_ts = np.zeros(tlim,dtype=DTYPE2)
    
    # the loop
    for t in xrange(tlim):
        amp = spatial_ts[t]
        flicker = flicker_vec[t]
        if flicker != 0:
            m_ts[t] = m_amp[flicker-1] * amp
            p_ts[t] = p_amp[flicker-1] * amp
            
    return m_ts, p_ts

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_strf_timeseries(np.ndarray[DTYPE2_t, ndim=1] stim_ts,
                             np.ndarray[DTYPE2_t, ndim=2] m_resp,
                             np.ndarray[DTYPE2_t, ndim=2] p_resp,
                             np.ndarray[DTYPE_t, ndim=1] flicker_vec):
    # cdef's
    cdef int s,t
    cdef int slim = stim_ts.shape[0]
    cdef int tlim = m_resp.shape[0]

    # initialize output variable
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] m_ts = np.zeros(slim,dtype=DTYPE2)
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] p_ts = np.zeros(slim,dtype=DTYPE2)

    # the loop
    for s in xrange(slim):
        amp = stim_ts[s]
        for t in xrange(tlim):
                m_ts[s] += m_resp[t,flicker_vec[s]-1] * amp
                p_ts[s] += p_resp[t,flicker_vec[s]-1] * amp

    return m_ts, p_ts

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_strf_betas_timeseries(np.ndarray[DTYPE2_t, ndim=1] stim_ts,
                                    np.ndarray[DTYPE2_t, ndim=2] m_resp,
                                    np.ndarray[DTYPE2_t, ndim=2] p_resp,
                                    np.ndarray[DTYPE_t, ndim=1] flicker_vec,
                                    DTYPE2_t m_beta, DTYPE2_t p_beta):
    # cdef's
    cdef int s,t
    cdef int slim = stim_ts.shape[0]
    cdef int tlim = m_resp.shape[0]

    # initialize output variable
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] stim = np.zeros(slim,dtype=DTYPE2)
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] buff = np.zeros(slim,dtype=DTYPE2)

    # the loop
    for s in xrange(slim):
        amp = stim_ts[s]
        for t in xrange(tlim):
                stim[s] += m_resp[t,flicker_vec[s]-1] * m_beta * amp + p_resp[t,flicker_vec[s]-1] * p_beta * amp

    return stim

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_strf_beta_ratio_timeseries(np.ndarray[DTYPE2_t, ndim=1] stim_ts,
                                        np.ndarray[DTYPE2_t, ndim=2] m_resp,
                                        np.ndarray[DTYPE2_t, ndim=2] p_resp,
                                        np.ndarray[DTYPE_t, ndim=1] flicker_vec,
                                        DTYPE2_t beta, DTYPE2_t beta_ratio):
    # cdef's
    cdef int s,t
    cdef int slim = stim_ts.shape[0]
    cdef int tlim = m_resp.shape[0]
    cdef DTYPE2_t beta_2 = beta_ratio/beta

    # initialize output variable
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] stim = np.zeros(slim,dtype=DTYPE2)

    # the loop
    for s in xrange(slim):
        amp = stim_ts[s]
        for t in xrange(tlim):
                stim[s] += m_resp[t,flicker_vec[s]-1] * beta * amp + p_resp[t,flicker_vec[s]-1] * beta_2 * amp

    return stim

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_strf_weight_timeseries(np.ndarray[DTYPE2_t, ndim=1] stim_ts,
                                    np.ndarray[DTYPE2_t, ndim=2] m_resp,
                                    np.ndarray[DTYPE2_t, ndim=2] p_resp,
                                    np.ndarray[DTYPE_t, ndim=1] flicker_vec,
                                    DTYPE2_t weight):
    # cdef's
    cdef int s,t
    cdef int slim = stim_ts.shape[0]
    cdef int tlim = m_resp.shape[0]

    # initialize output variable
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] stim = np.zeros(slim,dtype=DTYPE2)
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] buff = np.zeros(slim,dtype=DTYPE2)

    # the loop
    for s in xrange(slim):
        amp = stim_ts[s]
        for t in xrange(tlim):
                stim[s] += m_resp[t,flicker_vec[s]-1] * (1-weight) * amp + p_resp[t,flicker_vec[s]-1] * weight * amp

    return stim

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_rf_timeseries_nomask(np.ndarray[DTYPE3_t, ndim=3] stim_arr,
                                  np.ndarray[DTYPE2_t, ndim=2] rf):

    # cdef's
    cdef int i,j,k
    cdef int xlim = stim_arr.shape[0]
    cdef int ylim = stim_arr.shape[1]
    cdef int zlim = stim_arr.shape[2]

    # initialize output variable
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] stim = np.zeros(zlim,dtype=DTYPE2)

    # the loop
    for i in xrange(xlim):
        for j in xrange(ylim):
            for k in xrange(zlim):
                stim[k] += stim_arr[i,j,k]*rf[i,j]

    return stim

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_rf_timeseries(np.ndarray[DTYPE3_t, ndim=3] stim_arr,
                           np.ndarray[DTYPE2_t, ndim=2] rf,
                           np.ndarray[DTYPE_t, ndim=2] mask):

    # cdef's
    cdef int i,j,k
    cdef int xlim = stim_arr.shape[0]
    cdef int ylim = stim_arr.shape[1]
    cdef int zlim = stim_arr.shape[2]

    # initialize output variable
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] stim = np.zeros(zlim,dtype=DTYPE2)

    # the loop
    for i in xrange(xlim):
        for j in xrange(ylim):
            if mask[i,j] == 1:
                for k in xrange(zlim):
                    stim[k] += stim_arr[i,j,k]*rf[i,j]

    return stim

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_rf_timeseries_1D(np.ndarray[DTYPE2_t, ndim=2] stim_arr,
                              np.ndarray[DTYPE2_t, ndim=1] rf,
                              np.ndarray[DTYPE_t, ndim=1] mask):

    # cdef's
    cdef int i,j,k
    cdef int xlim = stim_arr.shape[0]
    cdef int ylim = stim_arr.shape[1]

    # initialize output variable
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] stim = np.zeros(ylim,dtype=DTYPE2)

    # the loop
    for i in xrange(xlim):
        if mask[i] == 1:
            for j in xrange(ylim):
                stim[j] += stim_arr[i,j]*rf[i]

    return stim

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_dog_timeseries(np.ndarray[DTYPE2_t, ndim=2] deg_x,
                                 np.ndarray[DTYPE2_t, ndim=2] deg_y,
                                 np.ndarray[DTYPE_t, ndim=3] stim_arr,
                                 DTYPE2_t x, DTYPE2_t y,
                                 DTYPE2_t sigma_center,
                                 DTYPE2_t sigma_surround):

    """
    Generate a time-series given a stimulus array and Gaussian parameters.

    Parameters
    ----------
    deg_x : 2D array
            The coordinate matrix along the horizontal dimension of the display (degrees)
    deg_y : 2D array
            The coordinate matrix along the vertical dimension of the display (degrees)
    x : float
       The x coordinate of the center of the Gaussian (degrees)
    y : float
       The y coordinate of the center of the Gaussian (degrees)
    s : float
       The dispersion of the Gaussian (degrees)
    beta : float
        The amplitude of the Gaussian

    Returns

    stim : ndarray
        The 1D array containing the stimulus energies given the Gaussian coordinates

    """

    # cdef's
    cdef int i,j,k,num_center_pixels, num_surround_pixels

    cdef DTYPE2_t sigma_center_factor2 = (2.0*sigma_center**2)
    cdef DTYPE2_t sigma_center_factor3 = (3.0*sigma_center)**2

    cdef DTYPE2_t sigma_surround_factor2 = (2.0*sigma_surround**2)
    cdef DTYPE2_t sigma_surround_factor3 = (3.0*sigma_surround)**2

    cdef int xlim = deg_x.shape[0]
    cdef int ylim = deg_x.shape[1]
    cdef int zlim = stim_arr.shape[2]
    cdef int add_center = 0
    cdef int add_surround = 0

    cdef DTYPE2_t d, gauss1D_center, gauss1D_surround
    cdef DTYPE2_t sum_gauss_center = 0.0
    cdef DTYPE2_t sum_gauss_surround = 0.0


    # initialize output variable
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] stim_center = np.zeros(zlim,dtype=DTYPE2)
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] stim_surround = np.zeros(zlim,dtype=DTYPE2)


    # the loop
    for i in xrange(xlim):
        for j in xrange(ylim):

            # find out the distance of the pixel from the prf
            d = (deg_x[i,j]-x)**2 + (deg_y[i,j]-y)**2

            # if its inside the surround ...
            if d <= sigma_surround_factor3:
                gauss1D_surround = exp(-d/sigma_surround_factor2)
                sum_gauss_surround += gauss1D_surround
                add_surround = 1
            else:
                add_surround = 0

            # if its inside the center ...
            if d <= sigma_center_factor3:
                gauss1D_center = exp(-d/sigma_center_factor2)
                sum_gauss_center += gauss1D_center
                add_center = 1
            else:
                add_center = 0

            # filter the pixel timeseries by the gauss
            for k in xrange(zlim):
                if add_center == 1:
                    stim_center[k] += stim_arr[i,j,k]*gauss1D_center

                if add_surround == 1:
                    stim_surround[k] += stim_arr[i,j,k]*gauss1D_surround


    # scale it by the integral
    stim_center /= sum_gauss_center
    stim_surround /= sum_gauss_surround

    return stim_center, stim_surround

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_og_timeseries(np.ndarray[DTYPE2_t, ndim=2] deg_x,
                           np.ndarray[DTYPE2_t, ndim=2] deg_y,
                           np.ndarray[DTYPE_t, ndim=3] stim_arr,
                           DTYPE2_t x, DTYPE2_t y, DTYPE2_t s):

    """
    Generate a time-series given a stimulus array and Gaussian parameters.

    Parameters
    ----------
    deg_x : 2D array
            The coordinate matrix along the horizontal dimension of the display (degrees)
    deg_y : 2D array
            The coordinate matrix along the vertical dimension of the display (degrees)
    x : float
       The x coordinate of the center of the Gaussian (degrees)
    y : float
       The y coordinate of the center of the Gaussian (degrees)
    s : float
       The dispersion of the Gaussian (degrees)
    beta : float
        The amplitude of the Gaussian

    Returns

    stim : ndarray
        The 1D array containing the stimulus energies given the Gaussian coordinates

    """

    # cdef's
    cdef int i,j,k,num_pixels_used
    cdef int xlim = deg_x.shape[0]
    cdef int ylim = deg_x.shape[1]
    cdef int zlim = stim_arr.shape[2]
    cdef DTYPE2_t pi = 3.14159265
    cdef DTYPE2_t gauss_area = 2.0*s**2
    cdef DTYPE2_t gauss_distance = (3.0*s)**2
    cdef DTYPE2_t gauss_integral = 2.0*pi*s**2
    cdef DTYPE2_t d, gauss1D
    cdef DTYPE2_t sum_gauss = 0.0

    # initialize output variable
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] stim = np.zeros(zlim,dtype=DTYPE2)

    # the loop
    for i in xrange(xlim):
        for j in xrange(ylim):
            d = (deg_x[i,j]-x)**2 + (deg_y[i,j]-y)**2
            if d <= gauss_distance:
                gauss1D = exp(-d/gauss_area)
                sum_gauss += gauss1D
                for k in xrange(zlim):
                    stim[k] += stim_arr[i,j,k]*gauss1D

    return stim

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_spectrotemporal_timeseries(np.ndarray[DTYPE2_t, ndim=1] freqs,
                                        np.ndarray[DTYPE2_t, ndim=2] spectrogram,
                                        DTYPE2_t center_freq, DTYPE2_t sd):

    """
    Generate a time-series given a stimulus array and Gaussian parameters.

    Parameters
    ----------
    deg_x : 2D array
            The coordinate matrix along the horizontal dimension of the display (degrees)
    deg_y : 2D array
            The coordinate matrix along the vertical dimension of the display (degrees)
    x : float
       The x coordinate of the center of the Gaussian (degrees)
    y : float
       The y coordinate of the center of the Gaussian (degrees)
    s : float
       The dispersion of the Gaussian (degrees)
    beta : float
        The amplitude of the Gaussian

    Returns

    stim : ndarray
        The 1D array containing the stimulus energies given the Gaussian coordinates

    """

    # cdef's
    cdef int i,j
    cdef DTYPE2_t s_factor2 = (2.0*sd**2)
    cdef DTYPE2_t s_factor3 = (3.0*sd)**2
    cdef int xlim = spectrogram.shape[0]
    cdef int ylim = spectrogram.shape[1]
    cdef DTYPE2_t d, gauss1D
    cdef DTYPE2_t sum_gauss = 0.0

    # initialize output variable
    cdef np.ndarray[DTYPE2_t,ndim=1,mode='c'] stim = np.zeros(ylim,dtype=DTYPE2)

    # the loop
    for i in xrange(xlim):

        d = (freqs[i]-center_freq)**2

        for j in xrange(ylim):

            if d <= s_factor3:

                #compute gauss
                gauss1D = exp(-d/s_factor2)

                # filter spectrogram
                stim[j] += spectrogram[i,j]*gauss1D

    return stim

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_2dcos_receptive_field(DTYPE2_t x, DTYPE2_t y, DTYPE2_t sigma, DTYPE2_t power,
                                np.ndarray[DTYPE2_t, ndim=2] deg_x,
                                np.ndarray[DTYPE2_t, ndim=2] deg_y):
    """
    Generate a Gaussian.

    Parameters
    ----------
    x : float
       The x coordinate of the center of the Gaussian (degrees)
    y : float
       The y coordinate of the center of the Gaussian (degrees)
    s : float
       The dispersion of the Gaussian (degrees)
    power : float
        The amplitude of the Gaussian
    deg_x : 2D array
            The coordinate matrix along the horizontal dimension of the display (degrees)
    deg_y : 2D array
            The coordinate matrix along the vertical dimension of the display (degrees)
            
            
    Returns
    
    stim : ndarray
        The 1D array containing the stimulus energies given the Gaussian coordinates
        
    """
    
    # cdef's
    cdef int i,j,k
    cdef int xlim = deg_x.shape[0]
    cdef int ylim = deg_x.shape[1]
    cdef DTYPE2_t d
    cdef DTYPE2_t pi = 3.14159265
    
    
    # initialize output variable
    cdef np.ndarray[DTYPE2_t, ndim=2, mode='c'] rf = np.zeros((xlim,ylim),dtype=DTYPE2)
    
    # the loop
    for i in xrange(xlim):
        for j in xrange(ylim):
            d = ((deg_x[i,j]-x)**2 + (deg_y[i,j]-y)**2)**0.5
            rf[i,j] = ((0.5*(1 + cos(d*pi/sigma) ))**power) * (d<=sigma)
            
    return rf

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_og_receptive_field(DTYPE2_t x, DTYPE2_t y, DTYPE2_t sigma,
                                np.ndarray[DTYPE2_t, ndim=2] deg_x,
                                np.ndarray[DTYPE2_t, ndim=2] deg_y):
    """
    Generate a Gaussian.

    Parameters
    ----------
    x : float
       The x coordinate of the center of the Gaussian (degrees)
    y : float
       The y coordinate of the center of the Gaussian (degrees)
    s : float
       The dispersion of the Gaussian (degrees)
    beta : float
        The amplitude of the Gaussian
    deg_x : 2D array
            The coordinate matrix along the horizontal dimension of the display (degrees)
    deg_y : 2D array
            The coordinate matrix along the vertical dimension of the display (degrees)


    Returns

    stim : ndarray
        The 1D array containing the stimulus energies given the Gaussian coordinates

    """

    # cdef's
    cdef int i,j,k
    cdef DTYPE2_t s_factor2 = (2.0*sigma**2)
    cdef int xlim = deg_x.shape[0]
    cdef int ylim = deg_x.shape[1]
    cdef DTYPE2_t d, gauss1D
    cdef DTYPE2_t pi = 3.14159265


    # initialize output variable
    cdef np.ndarray[DTYPE2_t, ndim=2, mode='c'] rf = np.zeros((xlim,ylim),dtype=DTYPE2)

    # the loop
    for i in xrange(xlim):
        for j in xrange(ylim):
            d = (deg_x[i,j]-x)**2 + (deg_y[i,j]-y)**2
            rf[i,j] = exp(-d/s_factor2)

    return rf

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_gabor_receptive_field(DTYPE2_t x0,
                                   DTYPE2_t y0,
                                   DTYPE2_t s0,
                                   DTYPE2_t theta,
                                   DTYPE2_t phi,
                                   DTYPE2_t cpd,
                                   np.ndarray[DTYPE2_t, ndim=2] deg_x,
                                   np.ndarray[DTYPE2_t, ndim=2] deg_y):

    # cdef's
    cdef int i,j,k
    cdef DTYPE2_t s_factor2 = (2.0*s0**2)
    cdef DTYPE2_t s_factor3 = (3.0*s0)**2
    cdef int xlim = deg_x.shape[0]
    cdef int ylim = deg_y.shape[1]
    cdef DTYPE2_t pi_180 = np.pi/180
    cdef DTYPE2_t theta_rad = theta * pi_180
    cdef DTYPE2_t phi_rad = phi * pi_180
    cdef DTYPE2_t pi_2 = np.pi * 2
    cdef DTYPE2_t d

    # fodder
    cdef np.ndarray[DTYPE2_t, ndim=2, mode='c'] XYt = np.zeros((xlim,ylim),dtype=DTYPE2)
    cdef np.ndarray[DTYPE2_t, ndim=2, mode='c'] XYf = np.zeros((xlim,ylim),dtype=DTYPE2)
    cdef np.ndarray[DTYPE2_t, ndim=2, mode='c'] grating = np.zeros((xlim,ylim),dtype=DTYPE2)
    cdef np.ndarray[DTYPE2_t, ndim=2, mode='c'] gauss = np.zeros((xlim,ylim),dtype=DTYPE2)
    cdef np.ndarray[DTYPE2_t, ndim=2, mode='c'] gabor = np.zeros((xlim,ylim),dtype=DTYPE2)

    # the loop
    for i in xrange(xlim):
        for j in xrange(ylim):

            # only compute inside the sigma*3
            d = (deg_x[i,j]-x0)**2 + (deg_y[i,j]-y0)**2

            if d <= s_factor3:

                # creating the grating
                XYt[i,j] =  (deg_x[i,j] * cos(theta_rad)) + (deg_y[i,j] * sin(theta_rad))
                XYf[i,j] = XYt[i,j] * cpd * pi_2
                grating[i,j] = sin(XYf[i,j] + phi_rad)

                # create the gaussian
                gauss[i,j] =  exp(-d/s_factor2)

                # create the gabor
                gabor[i,j] = gauss[i,j] * grating[i,j]

    return gabor

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_gabor_receptive_field(DTYPE2_t x0,
                                   DTYPE2_t y0,
                                   DTYPE2_t s0,
                                   DTYPE2_t theta,
                                   DTYPE2_t phi,
                                   DTYPE2_t cpd,
                                   np.ndarray[DTYPE2_t, ndim=2] deg_x,
                                   np.ndarray[DTYPE2_t, ndim=2] deg_y):

    # cdef's
    cdef int i,j,k
    cdef DTYPE2_t s_factor2 = (2.0*s0**2)
    cdef DTYPE2_t s_factor3 = (3.0*s0)**2
    cdef int xlim = deg_x.shape[0]
    cdef int ylim = deg_y.shape[1]
    cdef DTYPE2_t pi_180 = np.pi/180
    cdef DTYPE2_t theta_rad = theta * pi_180
    cdef DTYPE2_t phi_rad = phi * pi_180
    cdef DTYPE2_t pi_2 = np.pi * 2
    cdef DTYPE2_t d

    # fodder
    cdef np.ndarray[DTYPE2_t, ndim=2, mode='c'] XYt = np.zeros((xlim,ylim),dtype=DTYPE2)
    cdef np.ndarray[DTYPE2_t, ndim=2, mode='c'] XYf = np.zeros((xlim,ylim),dtype=DTYPE2)
    cdef np.ndarray[DTYPE2_t, ndim=2, mode='c'] grating = np.zeros((xlim,ylim),dtype=DTYPE2)
    cdef np.ndarray[DTYPE2_t, ndim=2, mode='c'] gauss = np.zeros((xlim,ylim),dtype=DTYPE2)
    cdef np.ndarray[DTYPE2_t, ndim=2, mode='c'] gabor = np.zeros((xlim,ylim),dtype=DTYPE2)

    # the loop
    for i in xrange(xlim):
        for j in xrange(ylim):

            # only compute inside the sigma*3
            d = (deg_x[i,j]-x0)**2 + (deg_y[i,j]-y0)**2

            if d <= s_factor3:

                # creating the grating
                XYt[i,j] =  (deg_x[i,j] * cos(theta_rad)) + (deg_y[i,j] * sin(theta_rad))
                XYf[i,j] = XYt[i,j] * cpd * pi_2
                grating[i,j] = sin(XYf[i,j] + phi_rad)

                # create the gaussian
                gauss[i,j] =  exp(-d/s_factor2)

                # create the gabor
                gabor[i,j] = gauss[i,j] * grating[i,j]

    return gabor

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_og_receptive_fields(
                                 np.ndarray[DTYPE2_t, ndim=1] xs,
                                 np.ndarray[DTYPE2_t, ndim=1] ys,
                                 np.ndarray[DTYPE2_t, ndim=1] ss,
                                 np.ndarray[DTYPE2_t, ndim=1] amps,
                                 np.ndarray[DTYPE2_t, ndim=2] deg_x,
                                 np.ndarray[DTYPE2_t, ndim=2] deg_y):
    # cdef's
    cdef int i,j,k
    cdef int xlim = deg_x.shape[0]
    cdef int ylim = deg_x.shape[1]
    cdef int zlim = len(xs)
    cdef DTYPE2_t s_factor2 = (2.0*ss[0]**2)

    # initilize the output variable
    cdef np.ndarray[DTYPE2_t, ndim=3, mode='c'] rfs = np.zeros((xlim,ylim,zlim),dtype=DTYPE2)

    for i in xrange(xlim):
        for j in xrange(ylim):
            for k in xrange(zlim):
                s_factor2 = (2.0*ss[k]**2)
                d = (deg_x[i,j]-xs[k])**2 + (deg_y[i,j]-ys[k])**2
                rfs[i,j,k] += exp(-d/s_factor2) * amps[k]

    return rfs



"""

First pass at a stimulus model for abstracting the qualities and functionality of a stimulus
into an abstract class.  For now, we'll assume the stimulus model only pertains to visual 
stimuli on a visual display over time (i.e., 3D).  Hopefully this can be extended to other stimuli
with an arbitrary number of dimensions (e.g., auditory stimuli).

"""
from __future__ import division
import ctypes

from scipy.signal import spectrogram
import numpy as np

from popeye.base import StimulusModel
from popeye.onetime import auto_attr
import popeye.utilities as utils

def generate_spectrogram(signal, Fs, tr_length, noverlap=0, bins_per_octave = 20*12, freq_min = 200 ,decibels=True):
    
    # window size is 1 TR x samples per seconds
    win = Fs*tr_length
    
    # find num freqs
    nfft=win
    
    # get spectrum
    freqs, times, spec = spectrogram(signal, Fs, nperseg=int(win), noverlap=noverlap, nfft=nfft)
    
    if decibels: # pragma: no cover
        
        # Ratio between adjacent frequencies in log-f axis
        fratio = 2**(1/bins_per_octave)
        
        # How manp.ny bins in log-f axis
        nbins = np.floor( np.log((Fs/2)/freq_min) / np.log(fratio) ).astype(int)
        
        # Freqs corresponding to each bin in FFT
        fftfrqs = np.arange(nfft/2)*(Fs/nfft)
        nfftbins = np.int(nfft/2);
        
        # Freqs corresponding to each bin in log F output
        logffrqs = freq_min * np.exp(np.log(2)*np.arange(nbins)/bins_per_octave);
        
        # Bandwidths of each bin in log F
        logfbws = logffrqs * (fratio - 1)
        
        ovfctr = 0.5475;   # Adjusted by hand to make sum(mx'*mx) close to 1.0
        
        # Weighting matrix mapping energy in FFT bins to logF bins
        # is a set of Gaussian profiles depending on the difference in 
        # frequencies, scaled by the bandwidth of that bin
        A = np.tile(logffrqs,(nfftbins,1)).T
        B = np.tile(fftfrqs,(nbins,1))
        C = np.tile(ovfctr*logfbws,(nfftbins,1)).T
        freqdiff = ( A - B )/C;
        
        
        # % Normalize rows by sqrt(E), so multiplying by mx' gets approx orig spectrum back
        mx = np.exp( -0.5*freqdiff**2 );
        D = np.sqrt(2*np.sum(mx**2,1))
        E = mx / np.tile(D, (nfftbins,1)).T
        Sx = spec.copy()
        Px = np.matrix(spec[1::])
        spec = np.array(np.matrix(mx) * Px)
        
        # output
        times = np.arange(spec.shape[-1])
        freqs = np.log2(logffrqs)/np.log2(10)
        
    return spec, freqs, times
        
class AuditoryStimulus(StimulusModel):
    
    
    def __init__(self, signal, Fs, tr_length, dtype):
        
        r"""A child of the StimulusModel class for auditory stimuli.
        
        Paramaters
        ----------
        
        
        dtype : string
            Sets the data type the stimulus array is cast into.
        
        tr_length : float
            The repetition time (TR) in seconds.
            
        """
        
        # this is a weird notation
        StimulusModel.__init__(self, signal, dtype, tr_length)
        
        # absorb the vars
        self.Fs = Fs
        self.tr_length = tr_length
        
        # # create spectrogram
        specgram, freqs, times, = generate_spectrogram(self.stim_arr, Fs, tr_length)
        
        # share them
        self.spectrogram = utils.generate_shared_array(specgram, ctypes.c_double)
        self.freqs = utils.generate_shared_array(freqs, ctypes.c_double)
        self.times = utils.generate_shared_array(times, ctypes.c_int16)



#!/usr/bin/python

""" Classes and functions for fitting population encoding models """

from __future__ import division
import warnings
warnings.simplefilter("ignore")

import numpy as np
from scipy.signal import fftconvolve
from scipy.interpolate import interp1d
from scipy.stats import linregress
import nibabel

from popeye.onetime import auto_attr
import popeye.utilities as utils
from popeye.base import PopulationModel, PopulationFit
from popeye.spinach import generate_rf_timeseries_1D

class AuditoryModel(PopulationModel):
    
    def __init__(self, stimulus, hrf_model, normalizer=utils.percent_change):
        
        r"""A 1D Gaussian population receptive field model [1]_.
        
        Paramaters
        ----------
        
        stimulus : `AuditoryStimulus` class object
            A class instantiation of the `AuditoryStimulus` class
            containing a representation of the auditory stimulus.
            
        hrf_model : callable
            A function that generates an HRF model given an HRF delay.
            For more information, see `popeye.utilties.double_gamma_hrf_hrf`
        
        
        References
        ----------
        
        .. [1] Thomas JM, Huber E, Stecker GC, Boynton GM, Saenz M, Fine I. (2015) 
        Population receptive field estimates in human auditory cortex. 
        NeuroImage 105, 428-439.
        
        """
        
        # invoke the base class
        PopulationModel.__init__(self, stimulus, hrf_model, normalizer)
    
    def generate_ballpark_prediction(self, center_freq, sigma):

        r"""
        Generate a prediction for the 1D Gaussian model.

        This function generates a prediction of the 1D Gaussian model, 
        given a stimulus and the stimulus-referred model parameters.  
        This function operates on a spectrogram representation of an 
        auditory stimulus. This method does not estimate the scaling
        paramter `beta` or the offset parameter `baseline`, since this
        method will be used for a grid-fit and these values can simply
        be calculated for a particular `center_freq` and `sigma` pair.
        
        Paramaters
        ----------
        
        center_freq : float
            The center frequency of the 1D Gaussian, units are in Hz.
            
        sigma : float
            The dispersion of the 1D Gaussian, units are in Hz.
            
        """
        
        # receptive field
        rf = np.exp(-((10**self.stimulus.freqs-10**center_freq)**2)/(2*(10**sigma)**2))
        rf /= (10**sigma*np.sqrt(2*np.pi))
        
        # evaluate entire RF
        mask = np.ones_like(rf).astype('uint8')
        
        # extract the response
        response = generate_rf_timeseries_1D(self.stimulus.spectrogram, rf, mask)
        
        # convolve it with the stimulus
        model = fftconvolve(response, self.hrf())[0:len(response)]
        
        # units
        model = self.normalizer(model)
        
        # regress out mean and amplitude
        beta, baseline = self.regress(model, self.data)
        
        # scale
        model *= beta
        
        # offset
        model += baseline
        
        return model
        
    def generate_prediction(self, center_freq, sigma, beta, baseline, unscaled=False):
        
        r"""
        Generate a prediction for the 1D Gaussian auditory pRF model.
        
        This function generates a prediction of the 1D Gaussian model, 
        given a stimulus and the stimulus-referred model parameters.  
        This function operates on a spectrogram representation of an 
        auditory stimulus.
        
        Paramaters
        ----------
        
        center_freq : float
            The center frequency of the 1D Gaussian, units are in Hz.
            
        sigma : float
            The dispersion of the 1D Gaussian, units are in Hz.
        
        baseline : float
            The y-intercept aka offset in amplitude to account for baseline.
        
        beta : float
            The scaling factor to account for arbitrary units of the BOLD signal.
        
        """ 
        
        # receptive field
        rf = np.exp(-((10**self.stimulus.freqs-10**center_freq)**2)/(2*(10**sigma)**2))
        rf /= (10**sigma*np.sqrt(2*np.pi))
        
        # evaluate entire RF
        mask = np.ones_like(rf).astype('uint8')
        
        # extract the response
        response = generate_rf_timeseries_1D(self.stimulus.spectrogram, rf, mask)
        
        # convolve it with the stimulus
        model = fftconvolve(response, self.hrf())[0:len(response)]
        
        # units
        model = self.normalizer(model)
        
        if unscaled:
            return model
        else:
            
            # scale it by beta
            model *= beta
            
            # offset
            model += baseline
            
            return model
        
class AuditoryFit(PopulationFit):
    
    def __init__(self, model, data, grids, bounds,
                 voxel_index=(1,2,3), Ns=None, auto_fit=True, verbose=0):
        
        r"""
        A class containing tools for fitting the 1D Gaussian auditory pRF model.
        
        The `AuditoryFit` class houses all the fitting tool that are associated with 
        estimatinga pRF model.  The `PopulationFit` takes a `AuditoryModel` instance 
        `model` and a time-series `data`.  In addition, extent and sampling-rate of a 
        brute-force grid-search is set with `grids` and `Ns`.  Use `bounds` to set 
        limits on the search space for each parameter.  
        
        Paramaters
        ----------
        
                
        model : `AuditoryModel` class instance
            An object representing the 1D Gaussian model.
        
        data : ndarray
            An array containing the measured BOLD signal of a single voxel.
        
        grids : tuple
            A tuple indicating the search space for the brute-force grid-search.
            The tuple contains pairs of upper and lower bounds for exploring a
            given dimension.  For example `grids=((-10,10),(0,5),)` will
            search the first dimension from -10 to 10 and the second from 0 to 5.
            These values cannot be `None`. 
            
            For more information, see `scipy.optimize.brute`.
        
        bounds : tuple
            A tuple containing the upper and lower bounds for each parameter
            in `parameters`.  If a parameter is not bounded, simply use
            `None`.  For example, `fit_bounds=((0,None),(-10,10),)` would 
            bound the first parameter to be any positive number while the
            second parameter would be bounded between -10 and 10.
        
        Ns : int
            Number of samples per stimulus dimension to sample during the ballpark search.
            
            For more information, see `scipy.optimize.brute`.
        
        voxel_index : tuple
            A tuple containing the index of the voxel being modeled. The 
            fitting procedure does not require a voxel index, but 
            collating the results across many voxels will does require voxel
            indices. With voxel indices, the brain volume can be reconstructed 
            using the newly computed model estimates.
        
        auto_fit : bool
            A flag for automatically running the fitting procedures once the 
            `GaussianFit` object is instantiated.
        
        verbose : int
            0 = silent
            1 = print the final solution of an error-minimization
            2 = print each error-minimization step
        
        """
        
        # invoke the base class
        PopulationFit.__init__(self, model, data, grids, bounds, 
                               voxel_index, Ns, auto_fit, verbose)
    
    @auto_attr
    def overloaded_estimate(self):
        
        return [10**self.center_freq, 10**self.sigma, self.beta, self.baseline]
    
    @auto_attr
    def center_freq0(self):
        return self.ballpark[0]
        
    @auto_attr
    def sigma0(self):
        return self.ballpark[1]
        
    @auto_attr
    def beta0(self):
        return self.ballpark[2]

    @auto_attr
    def baseline0(self):
        return self.ballpark[3]
    
    @auto_attr
    def center_freq(self):
        return self.estimate[0]
        
    @auto_attr
    def sigma(self):
        return self.estimate[1]
        
    @auto_attr
    def beta(self):
        return self.estimate[2]
    
    @auto_attr
    def baseline(self):
        return self.estimate[3]
    
    
    @auto_attr
    def center_freq_hz(self):
        return 10**self.center_freq
    
    @auto_attr
    def receptive_field(self):
        
        # generate stimulus time-series
        rf = np.exp(-((10**self.model.stimulus.freqs-10**self.center_freq)**2)/(2*(10**self.sigma)**2))
        rf /= (10**self.sigma*np.sqrt(2*np.pi))
        return rf
    
    @auto_attr
    def receptive_field_log10(self):
            
        # generate stimulus time-series
        rf = np.exp(-((self.model.stimulus.freqs-self.center_freq)**2)/(2*self.sigma**2))
        rf /= (self.sigma*np.sqrt(2*np.pi))
        return rf
       



#!/usr/bin/python

""" Classes and functions for fitting SpatioTemporal population receptive field models """

from __future__ import division, print_function, absolute_import
import time
import gc
import warnings
warnings.simplefilter("ignore")

import numpy as np
from scipy.stats import linregress
from scipy.signal import fftconvolve
from scipy.integrate import trapz, simps
from scipy.optimize import fmin
import nibabel

from popeye.onetime import auto_attr
import popeye.utilities as utils
from popeye.base import PopulationModel, PopulationFit
from popeye.spinach import generate_og_receptive_field, generate_mp_timeseries, generate_rf_timeseries

class SpatioTemporalModel(PopulationModel):
    
    """
    A Gaussian population receptive field model class
    
    """
    
    def __init__(self, stimulus, hrf_model, normalizer=utils.percent_change):
        
        """
        A spatiotemporal population receptive field model.
        
        Paramaters
        ----------
        
        stimulus : `VisualStimulus` class object
            A class instantiation of the `VisualStimulus` class
            containing a representation of the visual stimulus.
        
        
        """
        
        PopulationModel.__init__(self, stimulus, hrf_model, normalizer)
        
    # for the final solution, we use spatiotemporal
    def generate_ballpark_prediction(self, x, y, sigma, weight, hrf_delay):
        
        r"""
        Predict signal for the Gaussian Model using the downsampled stimulus.
        The rate of stimulus downsampling is defined in `model.stimulus.scale_factor`.
        
        Parameters
        __________
        x : float
            Horizontal location of the Gaussian RF.
        
        y: float 
            Vertical location of the Gaussian RF.
        
        sigma: float
            Dipsersion of the Gaussian RF.
        
        weight: float
            Mixture of the magnocellar and parvocellular temporal response to
            a flickering visual stimulus. The `weight` ranges between 0 and 1, 
            with 0 being a totally magnocellular response and ` being a totally
            parvocellular response.
        
        hrf_delay : float
            The delay of the peak and undershoot of the hemodynamic response
            function in seconds. This is a number varying about 0 which will
            be added to the 5 and 15, representing the constant delay of the
            peak and undershoot.
        
        """
        # mask for speed
        mask = self.distance_mask_coarse(x, y, sigma)
        
        # generate the RF
        spatial_rf = generate_og_receptive_field(x, y, sigma, self.stimulus.deg_x0, self.stimulus.deg_y0)
        spatial_rf /= ((2 * np.pi * sigma**2) * 1/np.diff(self.stimulus.deg_x0[0,0:2])**2)
        
        # spatial_response
        spatial_ts = generate_rf_timeseries(self.stimulus.stim_arr0, spatial_rf, mask)
        
        # temporal response
        m_ts, p_ts = generate_mp_timeseries(spatial_ts, self.m_amp, self.p_amp, self.stimulus.flicker_vec)
        
        # mix them
        mp_ts = (1-weight) * m_ts + weight * p_ts
        
        # convolve with HRF
        model = fftconvolve(mp_ts, self.hrf_model(hrf_delay, self.stimulus.tr_length))[0:len(mp_ts)]
        
        # units
        model = self.normalizer(model)
        
        # regress out mean and linear
        p = linregress(model, self.data)
        
        # scale
        model *= p[0]
        
        # offset
        model += p[1]
        
        return model
    
    # for the final solution, we use spatiotemporal
    def generate_prediction(self, x, y, sigma, weight, hrf_delay, beta, baseline, unscaled=False):
        
        r"""
        Predict signal for the Gaussian Model using the full resolution stimulus.
        
        Parameters
        __________
        x : float
            Horizontal location of the Gaussian RF.
        
        y: float 
            Vertical location of the Gaussian RF.
        
        sigma: float
            Dipsersion of the Gaussian RF.
        
        weight: float
            Mixture of the magnocellar and parvocellular temporal response to
            a flickering visual stimulus. The `weight` ranges between 0 and 1, 
            with 0 being a totally magnocellular response and ` being a totally
            parvocellular response.
        
        hrf_delay : float
            The delay of the peak and undershoot of the hemodynamic response
            function in seconds. This is a number varying about 0 which will
            be added to the 5 and 15, representing the constant delay of the
            peak and undershoot.
        
        beta : float
            Amplitude scaling factor to account for units.
            
        baseline: float
            Amplitude intercept to account for baseline.
        
        """
        
        # mask for speed
        mask = self.distance_mask(x, y, sigma)
        
        # generate the RF
        spatial_rf = generate_og_receptive_field(x, y, sigma, self.stimulus.deg_x, self.stimulus.deg_y)
        spatial_rf /= ((2 * np.pi * sigma**2) * 1/np.diff(self.stimulus.deg_x[0,0:2])**2)
        
        # spatial response
        spatial_ts = generate_rf_timeseries(self.stimulus.stim_arr, spatial_rf, mask)
        
        # temporal response
        m_ts, p_ts = generate_mp_timeseries(spatial_ts, self.m_amp, self.p_amp, self.stimulus.flicker_vec)
        
        # mix them
        mp_ts = (1-weight) * m_ts + weight * p_ts 
        
        # convolve with HRF
        model = fftconvolve(mp_ts, self.hrf_model(hrf_delay, self.stimulus.tr_length))[0:len(mp_ts)]
        
        # convert units
        model = self.normalizer(model)
        
        if unscaled:
            return model
        else:
            
            # scale
            model *= beta
            
            # offset
            model += baseline
            
            return model
    
    @auto_attr
    def p(self):
        
        r"""Returns the parvocellular sustained receptive field. The temporal dispersion
        parameter `tau` must be hard-set in the `SpatioTemporalModel.tau` by the user."""
        
        p = np.exp(-((self.t-self.center)**2)/(2*self.tau**2))
        p = p * 1/(np.sqrt(2*np.pi)*self.tau)
        return p
    
    @auto_attr
    def m(self):
        
        r"""Returns the magnocellar transient receptive field. The temporal dispersion
        parameter `tau` must be hard-set in the `SpatioTemporalModel.tau` by the user."""
        
        m = np.insert(np.diff(self.p),0,0)
        m = m/(simps(np.abs(m),self.t))
        return m
    
    def p_rf(self, tau):
        
        r""" Returns the sustained parvocellular receptive field. The sustained receptive field
        is a one dimensional gaussian in time.
        
        Parameters
        ----------
        
        tau : float
            The temporal dispersion of the sustained temporal receptive field.
            
        Returns
        -------
        
        p : ndarray
            The sustained parvocellular temporal receptive field normalized by its integral.
            
        """
        
        p = np.exp(-((self.t-self.center)**2)/(2*tau**2))
        p = p * 1/(np.sqrt(2*np.pi)*tau)
        return p
    
    def m_rf(self, tau):
        
        r""" Returns the transient magnocellular receptive field. The transient rectpive field
        is the temporal derivative of the sustained reponse.
        
        Parameters
        ----------
        
        tau : float
            The temporal dispersion of the sustained temporal receptive field.
            
        Returns
        -------
        
        m : ndarray
            The transient magnocellular temporal receptive field normalized by its integral.
            
        """
        
        p = self.p_rf(tau)
        m = np.insert(np.diff(p),0,0)
        m = m/(simps(np.abs(m),self.t))
        
        return m
    
    @auto_attr
    def t(self):
        
        r""" Returns the time coordinate."""
        
        return np.linspace(0, self.stimulus.tr_length, self.stimulus.fps * self.stimulus.tr_length)
    
    @auto_attr
    def center(self):
        
        r""" Returns the center of the time coordinate for constructing the temporal RFs."""
        
        return self.stimulus.tr_length/2
    
    @auto_attr
    def flickers(self):
        
        r""" Returns the temporal profiles of the flickering stimuli."""
        
        return np.sin(2 * np.pi * np.single(self.stimulus.flicker_hz) * self.t[:,np.newaxis])
    
    @auto_attr
    def m_resp(self):
        
        r""" Returns the transient magnocellular response to flicker stimulus."""
        
        m_resp = fftconvolve(self.flickers,self.m[:,np.newaxis])
        m_resp= utils.normalize(m_resp,-1,1)
        return m_resp
    
    def generate_m_resp(self, tau):
        
        r""" Returns the transient magnocellular response to flicker stimulus.
        
        Parameters
        ----------
        
        tau : float
            The temporal dispersion of the sustained temporal receptive field.
            
        Returns
        -------
        
        m_resp : ndarray
            The temporal response of the flickering visual stimulus convolved
            with the transient magnocellular temporal receptive field.
            
        """
        
        m_rf = self.m_rf(tau)
        m_resp = fftconvolve(self.flickers,m_rf[:,np.newaxis])
        m_resp = utils.normalize(m_resp,-1,1)
        return m_resp
    
    @auto_attr
    def m_amp(self):
        
        r""" Returns the amplitude of the transient mangmocellular response to flicker stimulus."""
        
        m_amp = np.sum(np.abs(self.m_resp),0)
        m_amp /= m_amp.max()
        return m_amp
    
    @auto_attr
    def p_resp(self):
        
        r""" Returns the sustained parvocellular response to flicker stimulus."""
        
        p_resp = fftconvolve(self.flickers,self.p[:,np.newaxis])
        p_resp = utils.normalize(p_resp,-1,1)
        return p_resp
    
    def generate_p_resp(self, tau):
        
        r""" Returns the sustained parvocellular response to flicker stimulus.
        
        Parameters
        ----------
        
        tau : float
            The temporal dispersion of the sustained temporal receptive field.
            
        Returns
        -------
        
        p_resp : ndarray
            The temporal response of the flickering visual stimulus convolved
            with the sustained parvocellular temporal receptive field.
            
        """
        
        p_rf = self.p_rf(tau)
        p_resp = fftconvolve(self.flickers,p_rf[:,np.newaxis])
        p_resp = utils.normalize(p_resp,-1,1)
        return p_resp
    
    @auto_attr
    def p_amp(self):
        
        r""" Returns the amplitude of the transient mangmocellular response to flicker stimulus."""
        
        p_amp = np.sum(np.abs(self.p_resp),0)
        p_amp /= np.max(p_amp)
        return p_amp
    
    def generate_receptive_field(self, x, y, sigma):
        
        r"""
        Generate a Gaussian receptive field in stimulus-referred coordinates.
        
        Parameters
        __________
        x : float
            Horizontal location of the Gaussian RF.
            
        y: float 
            Vertical location of the Gaussian RF.
            
        sigma: float
            Dipsersion of the Gaussian RF.
            
        """
        
        rf = generate_og_receptive_field(x, y, sigma, self.stimulus.deg_x, self.stimulus.deg_y)
        rf /= (2 * np.pi * sigma**2) * 1/np.diff(self.stimulus.deg_x[0,0:2])**2
        
        return rf
    
class SpatioTemporalFit(PopulationFit):
    
    """
    A spatiotemporal population receptive field fit class
    
    """
    
    def __init__(self, model, data, grids, bounds, 
                 voxel_index=(1,2,3), Ns=None, auto_fit=True, verbose=0):
        
        PopulationFit.__init__(self, model, data, grids, bounds, 
                               voxel_index, Ns, auto_fit, verbose)
    
    @auto_attr
    def overloaded_estimate(self):
        
        r""" Returns the user-defined overloaded estimate."""
        
        return [self.theta, self.rho, self.sigma, self.weight, self.hrf_delay, self.beta, self.baseline]
    
    @auto_attr
    def x0(self):
        return self.ballpark[0]
        
    @auto_attr
    def y0(self):
        return self.ballpark[1]
    
    @auto_attr
    def sigma0(self):
        return self.ballpark[2]
    
    @auto_attr
    def weight0(self):
        return self.ballpark[3]
        
    @auto_attr
    def hrf0(self):
        return self.ballpark[4]
        
    @auto_attr
    def beta0(self):
        return self.ballpark[5]

    @auto_attr
    def baseline0(self):
        return self.ballpark[6]
            
    @auto_attr
    def x(self):
        return self.estimate[0]
    
    @auto_attr
    def y(self):
        return self.estimate[1]
    
    @auto_attr
    def sigma(self):
        return self.estimate[2]
        
    @auto_attr
    def weight(self):
        return self.estimate[3]
    
    @auto_attr
    def hrf_delay(self):
        return self.estimate[4]
    
    @auto_attr
    def beta(self):
        return self.estimate[5]
    
    @auto_attr
    def baseline(self):
        return self.estimate[6]
    
    @auto_attr
    def rho(self):
        
        r""" Returns the eccentricity of the fitted pRF. """
        
        return np.sqrt(self.x**2+self.y**2)
    
    @auto_attr
    def theta(self):
        
        r""" Returns the polar angle of the fitted pRF. """
        
        return np.mod(np.arctan2(self.y,self.x),2*np.pi)
    
    @auto_attr
    def receptive_field(self):
        
        r""" Returns the fitted Gaussian pRF. """
        
        return self.model.generate_receptive_field(self.x, self.y, self.sigma)
    



import os
from itertools import repeat
import ctypes

import numpy as np
import numpy.testing as npt
import nose.tools as nt
from scipy.integrate import simps

import popeye.utilities as utils
from popeye import spatiotemporal_hrf as strf
from popeye.visual_stimulus import VisualStimulus, simulate_bar_stimulus, resample_stimulus
from popeye.spinach import generate_og_receptive_field

def test_strf_hrf_fit():
    
    viewing_distance = 38
    screen_width = 25
    thetas = np.tile(np.arange(0,360,90),2)
    thetas = np.insert(thetas,0,-1)
    thetas = np.append(thetas,-1)
    num_blank_steps = 20
    num_bar_steps = 20
    ecc = 10
    tr_length = 1.0
    frames_per_tr = 1.0
    scale_factor = 0.50
    pixels_down = 100
    pixels_across = 100
    dtype = ctypes.c_int16
    Ns = 3
    voxel_index = (1,2,3)
    auto_fit = True
    verbose = 1
    projector_hz = 480
    tau = 0.00875
    mask_size = 5
    
    # create the sweeping bar stimulus in memory
    stim = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, 
                                 screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
    
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(stim, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    stimulus.fps = projector_hz
    flicker_vec = np.zeros_like(stim[0,0,:]).astype('uint8')
    flicker_vec[1*20:5*20] = 1
    flicker_vec[5*20:9*20] = 2
    stimulus.flicker_vec = flicker_vec
    stimulus.flicker_hz = [10,20]
    
    # initialize the gaussian model
    model = strf.SpatioTemporalModel(stimulus, utils.spm_hrf)
    model.tau = tau
    model.mask_size = mask_size
    
    # generate a random pRF estimate
    x = -2.24
    y = 1.58
    sigma = 1.23
    weight = 0.90
    hrf_delay = -0.13
    beta = 0.88
    baseline = -0.25
    
    # create the "data"
    data =  model.generate_prediction(x, y, sigma, weight, hrf_delay, beta, baseline)
    
    # set search grid
    x_grid = utils.grid_slice(-8.0,7.0,4)
    y_grid = utils.grid_slice(-8.0,7.0,4)
    s_grid = utils.grid_slice(0.75,3.0,4)
    w_grid = utils.grid_slice(0.05,0.95,4)
    h_grid = utils.grid_slice(-0.25,0.25,4)
    
    # set search bounds
    x_bound = (-10,10)
    y_bound =  (-10,10)
    s_bound = (1/stimulus.ppd,10)
    w_bound = (1e-8,1.0)
    b_bound = (1e-8,1e5)
    u_bound = (None, None)
    h_bound = (-2.0,2.0)
    
    # loop over each voxel and set up a GaussianFit object
    grids = (x_grid, y_grid, s_grid, w_grid, h_grid,)
    bounds = (x_bound, y_bound, s_bound, w_bound, h_bound, b_bound, u_bound)
    
    # fit the response
    fit = strf.SpatioTemporalFit(model, data, grids, bounds)
    
    # coarse fit
    ballpark = [-3.0,
     2.0,
     1.5,
     0.95,
     -0.0833333,
     0.8992611,
     -0.25]
     
    npt.assert_almost_equal((fit.x0,fit.y0,fit.sigma0,fit.weight0,fit.hrf0,fit.beta0,fit.baseline0), ballpark)
    
    # fine fit
    npt.assert_almost_equal(fit.x, x)
    npt.assert_almost_equal(fit.y, y)
    npt.assert_almost_equal(fit.sigma, sigma)
    npt.assert_almost_equal(fit.weight, weight)
    npt.assert_almost_equal(fit.beta, beta)
    npt.assert_almost_equal(fit.baseline, baseline)
    
    # overloaded
    npt.assert_almost_equal(fit.overloaded_estimate, [ 2.5272803, 2.7411676, 1.23, 0.9, -0.13, 0.88, -0.25])
    
    m_rf = fit.model.m_rf(fit.model.tau)
    p_rf = fit.model.p_rf(fit.model.tau)
    npt.assert_almost_equal(simps(np.abs(m_rf)),simps(p_rf),5)
    
    # responses
    m_resp = fit.model.generate_m_resp(fit.model.tau)
    p_resp = fit.model.generate_p_resp(fit.model.tau)
    npt.assert_(np.max(m_resp,0)[0]<np.max(m_resp,0)[1])
    npt.assert_(np.max(p_resp,0)[0]>np.max(p_resp,0)[1])
    
    # amps
    npt.assert_(fit.model.m_amp[0]<fit.model.m_amp[1])
    npt.assert_(fit.model.p_amp[0]>fit.model.p_amp[1])
    
    # receptive field
    rf = generate_og_receptive_field(x, y, sigma, fit.model.stimulus.deg_x, fit.model.stimulus.deg_y)
    rf /= (2 * np.pi * sigma**2) * 1/np.diff(model.stimulus.deg_x[0,0:2])**2
    npt.assert_almost_equal(np.round(rf.sum()), np.round(fit.receptive_field.sum())) 
    
    # test model == fit RF
    npt.assert_almost_equal(np.round(fit.model.generate_receptive_field(x,y,sigma).sum()), np.round(fit.receptive_field.sum()))




import os
import sharedmem
import ctypes
import pickle
import time

import numpy as np
import numpy.testing as npt
import nose.tools as nt

import popeye.utilities as utils
import popeye.og as og
from popeye.visual_stimulus import VisualStimulus, simulate_bar_stimulus

def test_cache_model_slice():
    
    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,90)
    thetas = np.insert(thetas,0,-1)
    thetas = np.append(thetas,-1)
    num_blank_steps = 20
    num_bar_steps = 20
    ecc = 10
    tr_length = 1.5
    frames_per_tr = 1.0
    scale_factor = 0.50
    pixels_across = 100
    pixels_down = 100
    dtype = ctypes.c_int16
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, screen_width, 
                                thetas, num_bar_steps, num_blank_steps, ecc, clip=0.01)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.spm_hrf)
    model.hrf_delay = 0
    model.mask_size = 5
    
    # set cache grids
    x_grid = utils.grid_slice(-10, 10, 5)
    y_grid = utils.grid_slice(-10, 10, 5)
    s_grid = utils.grid_slice(0.55,5.25, 5)
    grids = (x_grid, y_grid, s_grid,)
    
    # seed rng
    np.random.seed(4932)
    
    # cache the pRF model
    cache = model.cache_model(grids, ncpus=3)
    
    # save it out
    pickle.dump(cache, open('/tmp/og_cached_model.pkl','wb'))
    
    # make sure its the right size
    cached_model = pickle.load(open('/tmp/og_cached_model.pkl','rb'))
    
    nt.assert_equal(np.sum([c[0] for c in cache]),np.sum([c[0] for c in cached_model]))

def test_cache_model_Ns():

    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,90)
    thetas = np.insert(thetas,0,-1)
    thetas = np.append(thetas,-1)
    num_blank_steps = 20
    num_bar_steps = 20
    ecc = 10
    tr_length = 1.5
    frames_per_tr = 1.0
    scale_factor = 0.50
    pixels_across = 100
    pixels_down = 100
    dtype = ctypes.c_int16
    Ns = 5
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, screen_width, 
                                thetas, num_bar_steps, num_blank_steps, ecc, clip=0.01)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.spm_hrf)
    model.hrf_delay = 0
    model.mask_size = 5
    
    # set cache grids
    x_grid = (-10, 10)
    y_grid = (-10, 10)
    s_grid = (0.55,5.25)
    grids = (x_grid, y_grid, s_grid,)
    
    # seed rng
    np.random.seed(4932)
    
    # cache the pRF model
    cache = model.cache_model(grids, Ns=Ns, ncpus=3)
    
    # save it out
    pickle.dump(cache, open('/tmp/og_cached_model.pkl','wb'))
    
    # make sure its the right size
    cached_model = pickle.load(open('/tmp/og_cached_model.pkl','rb'))
    
    nt.assert_equal(np.sum([c[0] for c in cache]),np.sum([c[0] for c in cached_model]))
    
    
def test_resurrect_model():
    
    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,90)
    thetas = np.insert(thetas,0,-1)
    thetas = np.append(thetas,-1)
    num_blank_steps = 20
    num_bar_steps = 20
    ecc = 10
    tr_length = 1.5
    frames_per_tr = 1.0
    scale_factor = 0.50
    pixels_across = 100
    pixels_down = 100
    dtype = ctypes.c_int16
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, screen_width, 
                                thetas, num_bar_steps, num_blank_steps, ecc, clip=0.01)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # set cache grids
    x_grid = utils.grid_slice(-10, 10, 5)
    y_grid = utils.grid_slice(-10, 10, 5)
    s_grid = utils.grid_slice(0.55,5.25, 5)
    grids = (x_grid, y_grid, s_grid,)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.spm_hrf)
    model.hrf_delay = 0
    model.mask_size = 5
    
    # seed rng
    np.random.seed(4932)
    
    # cache the model
    cache = model.cache_model(grids, ncpus=3)
    
    # resurrect cached model
    cached_model_path = '/tmp/og_cached_model.pkl'
    model = og.GaussianModel(stimulus, utils.double_gamma_hrf, cached_model_path=cached_model_path)
    model.hrf_delay = 0
    model.mask_size = 5
    
    # make sure the same
    nt.assert_true(np.sum([c[0] for c in cache] -  model.cached_model_timeseries) == 0)
    nt.assert_true(np.sum([c[1] for c in cache] -  model.cached_model_parameters) == 0)
    

def test_resurrect_model():
    
    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,90)
    thetas = np.insert(thetas,0,-1)
    thetas = np.append(thetas,-1)
    num_blank_steps = 20
    num_bar_steps = 20
    ecc = 10
    tr_length = 1.5
    frames_per_tr = 1.0
    scale_factor = 0.50
    pixels_across = 100
    pixels_down = 100
    dtype = ctypes.c_int16
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, screen_width, 
                                thetas, num_bar_steps, num_blank_steps, ecc, clip=0.01)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # set cache grids
    x_grid = utils.grid_slice(-10, 10, 5)
    y_grid = utils.grid_slice(-10, 10, 5)
    s_grid = utils.grid_slice(0.55,5.25, 5)
    grids = (x_grid, y_grid, s_grid,)
    
    # set search bounds
    x_bound = (-12.0,12.0)
    y_bound = (-12.0,12.0)
    s_bound = (0.001,12.0)
    b_bound = (1e-8,None)
    m_bound = (None,None)
    bounds = (x_bound, y_bound, s_bound, b_bound, m_bound)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.double_gamma_hrf)
    model.hrf_delay = 0
    model.mask_size = 5
    
    # seed rng
    np.random.seed(4932)
    
    # resurrect cached model
    cached_model_path = '/tmp/og_cached_model.pkl'
    model = og.GaussianModel(stimulus, utils.double_gamma_hrf, cached_model_path=cached_model_path)
    model.hrf_delay = 0
    model.mask_size = 5
    
    # pluck an estimate and create timeseries
    x, y, sigma = model.cached_model_parameters[50]
    beta = 1.25
    baseline = -0.25
    
    # create "data"
    data = model.generate_prediction(x, y, sigma, beta, baseline)
    
    # fit it
    fit = og.GaussianFit(model, data, grids, bounds)
    
    # assert
    nt.assert_true(np.all([x,y,sigma]==fit.ballpark))
    npt.assert_almost_equal(np.array([x,y,sigma,beta,baseline]),fit.estimate)


def test_resurrect_model():
    
    # stimulus features
    viewing_distance = 38
    screen_width = 25
    thetas = np.arange(0,360,90)
    thetas = np.insert(thetas,0,-1)
    thetas = np.append(thetas,-1)
    num_blank_steps = 20
    num_bar_steps = 20
    ecc = 10
    tr_length = 1.5
    frames_per_tr = 1.0
    scale_factor = 1.0
    pixels_across = 100
    pixels_down = 100
    dtype = ctypes.c_int16
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, screen_width, 
                                thetas, num_bar_steps, num_blank_steps, ecc, clip=0.01)
                                
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # set cache grids
    x_grid = utils.grid_slice(-10, 10, 5)
    y_grid = utils.grid_slice(-10, 10, 5)
    s_grid = utils.grid_slice(0.55,5.25, 5)
    grids = (x_grid, y_grid, s_grid,)
    
    # set search bounds
    x_bound = (-12.0,12.0)
    y_bound = (-12.0,12.0)
    s_bound = (0.001,12.0)
    b_bound = (1e-8,None)
    m_bound = (None,None)
    bounds = (x_bound, y_bound, s_bound, b_bound, m_bound)
    
    # initialize the gaussian model
    model = og.GaussianModel(stimulus, utils.double_gamma_hrf)
    model.hrf_delay = 0
    model.mask_size = 5
    
    cache = model.cache_model(grids, ncpus=3)
    
    # seed rng
    np.random.seed(4932)
    
    # pluck an estimate and create timeseries
    x, y, sigma = cache[51][1]
    beta = 1.25
    baseline = 0.25
    
    # create "data"
    data = cache[51][0]
    
    # fit it
    fit = og.GaussianFit(model, data, grids, bounds, verbose=0)
    
    # assert
    npt.assert_equal(fit.estimate,fit.ballpark)
    
    # create "data"
    data = model.generate_prediction(x,y,sigma,beta,baseline)
    
    # fit it
    fit = og.GaussianFit(model, data, grids, bounds, verbose=0)
    
    # assert
    npt.assert_almost_equal(np.sum(fit.scaled_ballpark_prediction-fit.data)**2,0)
    
    
    


import os, ctypes

import numpy as np
import numpy.testing as npt
import nose.tools as nt
from scipy.integrate import simps

import popeye.utilities as utils
from popeye import dog
from popeye.visual_stimulus import VisualStimulus, simulate_bar_stimulus

def test_dog():
    
    # stimulus features
    viewing_distance = 31
    screen_width = 41
    thetas = np.arange(0,360,90)
    # thetas = np.insert(thetas,0,-1)
    # thetas = np.append(thetas,-1)
    num_blank_steps = 0
    num_bar_steps = 30
    ecc = 10
    tr_length = 1.0
    frames_per_tr = 1.0
    scale_factor = 0.50
    pixels_down = 100
    pixels_across = 100
    dtype = ctypes.c_int16
    auto_fit = True
    verbose = 0
    
    # create the sweeping bar stimulus in memory
    bar = simulate_bar_stimulus(pixels_across, pixels_down, viewing_distance, 
                                screen_width, thetas, num_bar_steps, num_blank_steps, ecc)
    
    # create an instance of the Stimulus class
    stimulus = VisualStimulus(bar, viewing_distance, screen_width, scale_factor, tr_length, dtype)
    
    # initialize the gaussian model
    model = dog.DifferenceOfGaussiansModel(stimulus, utils.spm_hrf)
    model.hrf_delay = 0
    model.mask_size = 20
    
    # set the pRF params
    x = 2.2
    y = 2.5
    sigma = 0.90
    sigma_ratio = 1.5
    volume_ratio = 0.5
    beta = 0.25
    baseline = -0.10
    
    # create "data"
    data = model.generate_prediction(x, y, sigma, sigma_ratio, volume_ratio, beta, baseline)
    
    # set up the grids
    x_grid = utils.grid_slice(-5,5,4)
    y_grid = utils.grid_slice(-5,5,4)
    s_grid = utils.grid_slice(1/stimulus.ppd0*1.10,3.5,4)
    sr_grid = utils.grid_slice(1.0,2.0,4)
    vr_grid = utils.grid_slice(0.10,0.90,4)
    grids = (x_grid, y_grid, s_grid, sr_grid, vr_grid,)
    
    # set up the bounds
    x_bound = (-ecc,ecc)
    y_bound = (-ecc,ecc)
    s_bound = (1/stimulus.ppd,5)
    sr_bound = (1.0,None)
    vr_bound = (1e-8,1.0)
    bounds = (x_bound, y_bound, s_bound, sr_bound, vr_bound,)
    
    # fit it
    fit = dog.DifferenceOfGaussiansFit(model, data, grids, bounds)
    
    # coarse fit
    ballpark = [1.666666666666667,
                1.666666666666667,
                2.8243187483428391,
                1.9999999999999998,
                0.10000000000000001]
                
    npt.assert_almost_equal((fit.x0,fit.y0,fit.s0,fit.sr0,fit.vr0), ballpark)
    # the baseline/beta should be 0/1 when regressed data vs. estimate
    (m,b) = np.polyfit(fit.scaled_ballpark_prediction, data, 1)
    npt.assert_almost_equal(m, 1.0)
    npt.assert_almost_equal(b, 0.0)

    # fine fit
    npt.assert_almost_equal(fit.x, x, 2)
    npt.assert_almost_equal(fit.y, y, 2)
    npt.assert_almost_equal(fit.sigma, sigma, 2)
    npt.assert_almost_equal(fit.sigma_ratio, sigma_ratio, 1)
    npt.assert_almost_equal(fit.volume_ratio, volume_ratio, 1)
    
    # test the RF
    rf = fit.model.receptive_field(*fit.estimate[0:-2])
    est = fit.estimate[0:-2].copy()
    rf_new = fit.model.receptive_field(*est)
    value_1 = np.sqrt(simps(simps(rf))) 
    value_2 = np.sqrt(simps(simps(rf_new)))
    nt.assert_almost_equal(value_1, value_2)
    
    # polar coordinates
    npt.assert_almost_equal([fit.theta,fit.rho],[np.arctan2(y,x),np.sqrt(x**2+y**2)], 4)
    
